import json
import pathlib
from unittest.mock import patch

import pytest

from dissect.target.plugins.apps.webserver import iis
from dissect.target.plugins.os.windows import amcache
from dissect.target.tools.dump import run, state, utils
from tests._utils import absolute_path


@pytest.fixture
def target_win_iis_amcache(target_win_tzinfo, fs_win):
    config_path = absolute_path("_data/plugins/apps/webserver/iis/iis-applicationHost-iis.config")
    data_dir = absolute_path("_data/plugins/apps/webserver/iis/iis-logs-iis")

    fs_win.map_file("windows/system32/inetsrv/config/applicationHost.config", config_path)
    fs_win.map_dir("Users/John/iis-logs", data_dir)

    target_win_tzinfo.add_plugin(iis.IISLogsPlugin)

    amcache_file = absolute_path("_data/plugins/os/windows/amcache/amcache-new.hve")
    fs_win.map_file("windows/appcompat/programs/amcache.hve", amcache_file)

    target_win_tzinfo.add_plugin(amcache.AmcachePlugin)
    return target_win_tzinfo


@pytest.mark.parametrize(
    "serialization_name,compression_name",
    [
        ("jsonlines", "gzip"),
        ("jsonlines", "bzip2"),
        ("jsonlines", None),
        ("msgpack", "gzip"),
        ("msgpack", "bzip2"),
        ("msgpack", None),
    ],
)
def test_execute_pipeline(
    compression_name,
    serialization_name,
    target_win_iis_amcache,
    tmp_path,
):
    def mock_get_targets(_):
        yield target_win_iis_amcache

    def mock_log_progress(stream):
        count = 0
        for element in stream:
            yield element
            count += 1
        assert count == 128

    with patch("dissect.target.tools.dump.run.get_targets", new=mock_get_targets), patch(
        "dissect.target.tools.dump.run.log_progress", new=mock_log_progress
    ):
        output_dir = tmp_path / "output"

        serialization = utils.Serialization(serialization_name)
        compression = utils.Compression(compression_name)

        functions = "iis.logs,amcache.applications"

        run.execute_pipeline(
            targets=["dummy"],
            functions=functions,
            output_dir=output_dir,
            serialization=serialization,
            compression=compression,
        )

        target_name = target_win_iis_amcache.name

        # verify that iis records are in place

        assert (output_dir / target_name / "iis").exists()

        serialization_ext = utils.SERIALIZERS[serialization]["ext"]
        compression_ext = utils.COMPRESSION_TO_EXT[compression]

        iis_sink_filename = f"filesystem_windows_iis_logs.{serialization_ext}"
        if compression_ext:
            iis_sink_filename += f".{compression_ext}"

        assert (output_dir / target_name / "iis" / iis_sink_filename).exists()

        # verify that amcache.applications records are in place

        assert (output_dir / target_name / "amcache").exists()

        amcache_sink_filename = f"windows_appcompat_InventoryApplication.{serialization_ext}"
        if compression_ext:
            amcache_sink_filename += f".{compression_ext}"

        assert (output_dir / target_name / "amcache" / amcache_sink_filename).exists()

        # verify that serialized state is in place
        state_path = output_dir / state.STATE_FILE_NAME
        assert state_path.exists()

        with open(state_path, "r") as f:
            state_blob = json.loads(f.read())

        assert state_blob["compression"] == compression_name
        assert state_blob["serialization"] == serialization_name
        assert state_blob["target_paths"] == ["dummy"]
        assert state_blob["functions"] == functions

        sink_blobs = state_blob["sinks"]

        assert len(sink_blobs) == 2
        assert {s["func"] for s in sink_blobs} == set(functions.split(","))

        assert all(s["is_dirty"] is False for s in sink_blobs)

        # validate iis sink blob
        iis_sink_blob = next(s for s in state_blob["sinks"] if s["func"] == "iis.logs")
        assert iis_sink_blob["record_count"] == 10
        assert iis_sink_blob["target_path"] == str(target_win_iis_amcache.path)
        assert iis_sink_blob["path"] == str(pathlib.Path(target_name) / "iis" / iis_sink_filename)

        # validate amcache sink blob
        amcache_sink_blob = next(s for s in state_blob["sinks"] if s["func"] == "amcache.applications")
        assert amcache_sink_blob["record_count"] == 118
        assert amcache_sink_blob["target_path"] == str(target_win_iis_amcache.path)
        assert amcache_sink_blob["path"] == str(pathlib.Path(target_name) / "amcache" / amcache_sink_filename)


@pytest.mark.parametrize("limit", [5, 15, None])
def test_execute_pipeline_limited(limit, target_win_iis_amcache, tmp_path):
    def mock_get_targets(_):
        yield target_win_iis_amcache

    def mock_log_progress(stream):
        count = 0
        for element in stream:
            yield element
            count += 1

        if limit:
            assert count == limit
        else:
            assert count == 128

    with patch("dissect.target.tools.dump.run.get_targets", new=mock_get_targets), patch(
        "dissect.target.tools.dump.run.log_progress", new=mock_log_progress
    ):
        output_dir = tmp_path / "output"

        functions = "iis.logs,amcache.applications"

        run.execute_pipeline(
            targets=["dummy"],
            functions=functions,
            output_dir=output_dir,
            serialization=utils.Serialization.JSONLINES,
            limit=limit,
        )

        target_name = target_win_iis_amcache.name

        # verify that iis records are in place
        assert (output_dir / target_name / "iis").exists()

        serialization_ext = utils.SERIALIZERS[utils.Serialization.JSONLINES]["ext"]
        iis_sink_filename = f"filesystem_windows_iis_logs.{serialization_ext}"

        assert (output_dir / target_name / "iis" / iis_sink_filename).exists()

        # verify that serialized state is in place
        state_path = output_dir / state.STATE_FILE_NAME
        assert state_path.exists()

        with open(state_path, "r") as f:
            state_blob = json.loads(f.read())

        assert state_blob["compression"] is None
        sink_blobs = state_blob["sinks"]

        iis_sink_blob = next(s for s in state_blob["sinks"] if s["func"] == "iis.logs")

        # verify that amcache.applications records are absent if `limit` is smaller/equal than 10 (the number
        # of iis records), and present if `limit` is `None` or larger than 10.
        if limit and limit <= 10:
            assert not (output_dir / target_name / "amcache").exists()

            assert len(sink_blobs) == 1
            assert {s["func"] for s in sink_blobs} == {"iis.logs"}

            assert iis_sink_blob["record_count"] == limit
            if limit < 10:
                # if limit cuts the record stream, the sink should be marked as dirty
                assert iis_sink_blob["is_dirty"]
            else:
                assert not iis_sink_blob["is_dirty"]

        else:
            assert (output_dir / target_name / "amcache").exists()
            amcache_sink_filename = f"windows_appcompat_InventoryApplication.{serialization_ext}"
            assert (output_dir / target_name / "amcache" / amcache_sink_filename).exists()

            assert len(sink_blobs) == 2
            assert {s["func"] for s in sink_blobs} == set(functions.split(","))

            amcache_sink_blob = next(s for s in state_blob["sinks"] if s["func"] == "amcache.applications")

            if limit:
                assert amcache_sink_blob["record_count"] == limit - 10
                assert amcache_sink_blob["is_dirty"]
            else:
                assert amcache_sink_blob["record_count"] == 118
                assert not amcache_sink_blob["is_dirty"]



from __future__ import annotations

import logging
import re
import warnings
from dataclasses import dataclass
from datetime import datetime
from typing import Iterator, Optional

from dissect.target import Target
from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, export

warnings.simplefilter(action="ignore", category=FutureWarning)
log = logging.getLogger(__name__)

SchedLgURecord = TargetRecordDescriptor(
    "windows/tasks/log/schedlgu",
    [
        ("datetime", "ts"),
        ("string", "job"),
        ("string", "command"),
        ("string", "status"),
        ("uint32", "exit_code"),
        ("string", "version"),
    ],
)

JOB_REGEX_PATTERN = re.compile(r"\"(.*?)\" \((.*?)\)")
SCHEDLGU_REGEX_PATTERN = re.compile(r"\".+\n.+\n\s{4}.+\n|\".+\n.+", re.MULTILINE)


@dataclass(order=True)
class SchedLgU:
    ts: datetime = None
    job: str = None
    status: str = None
    command: str = None
    exit_code: int = None
    version: str = None

    @staticmethod
    def _sanitize_ts(ts: str) -> datetime:
        # sometimes "at" exists before the timestamp
        ts = ts.strip("at ")
        try:
            ts = datetime.strptime(ts, "%m/%d/%Y %I:%M:%S %p")
        except ValueError:
            ts = datetime.strptime(ts, "%d-%m-%Y %H:%M:%S")

        return ts

    @staticmethod
    def _parse_job(line: str) -> tuple[str, Optional[str]]:
        matches = JOB_REGEX_PATTERN.match(line)
        if matches:
            return matches.groups()

        log.warning("SchedLgU failed to parse job and command from line: '%s'. Returning line.", line)
        return line, None

    @classmethod
    def from_line(cls, line: str) -> SchedLgU:
        """Parse a group of SchedLgU.txt lines."""
        event = cls()
        lines = line.splitlines()

        # Events can have 2 or 3 lines as a group in total. An example of a complete task job event is:
        # "Symantec NetDetect.job" (NDETECT.EXE)
        #     Finished 14-9-2003 13:21:01
        #     Result: The task completed with an exit code of (65).
        if len(lines) == 3:
            event.job, event.command = cls._parse_job(lines[0])
            event.status, event.ts = lines[1].split(maxsplit=1)
            event.exit_code = int(lines[2].split("(")[1].rstrip(")."))

        # Events that have 2 lines as a group can be started task job event or the Task Scheduler Service. Examples:
        #   "Symantec NetDetect.job" (NDETECT.EXE)
        #        Started at 14-9-2003 13:26:00
        elif len(lines) == 2 and ".job" in lines[0]:
            event.job, event.command = cls._parse_job(lines[0])
            event.status, event.ts = lines[1].split(maxsplit=1)

        # Events without a task job event are the Task Scheduler Service events. Which can look like this:
        # "Task Scheduler Service"
        #      Exited at 14-9-2003 13:40:24
        # OR
        # "Task Scheduler Service"
        # 6.0.6000.16386 (vista_rtm.061101-2205)
        elif len(lines) == 2:
            event.job = lines[0].strip('"')

            if lines[1].startswith("\t") or lines[1].startswith(" "):
                event.status, event.ts = lines[1].split(maxsplit=1)
            else:
                event.version = lines[1]

        if event.ts:
            event.ts = cls._sanitize_ts(event.ts)

        return event


class SchedLgUPlugin(Plugin):
    """Plugin for parsing the Task Scheduler Service transaction log file (SchedLgU.txt)."""

    PATHS = {
        "sysvol/SchedLgU.txt",
        "sysvol/windows/SchedLgU.txt",
        "sysvol/windows/tasks/SchedLgU.txt",
        "sysvol/winnt/tasks/SchedLgU.txt",
    }

    def __init__(self, target: Target) -> None:
        self.target = target
        self.paths = [self.target.fs.path(path) for path in self.PATHS if self.target.fs.path(path).exists()]

    def check_compatible(self) -> None:
        if len(self.paths) == 0:
            raise UnsupportedPluginError("No SchedLgU.txt file found.")

    @export(record=SchedLgURecord)
    def schedlgu(self) -> Iterator[SchedLgURecord]:
        """Return all events in the Task Scheduler Service transaction log file (SchedLgU.txt).

        Older Windows systems may log ``.job`` tasks that get started remotely in the SchedLgU.txt file.
        In addition, this log file records when the Task Scheduler service starts and stops.

        Adversaries may use malicious ``.job`` files to gain persistence on a system.

        Yields SchedLgURecord with fields:

        .. code-block:: text

            ts (datetime): The timestamp of the event.
            job (str): The name of the .job file.
            command (str): The command executed.
            status (str): The status of the event (finished, completed, exited, stopped).
            exit_code (int): The exit code of the event.
            version (str): The version of the Task Scheduler service.
        """

        for path in self.paths:
            content = path.read_text(encoding="UTF-16", errors="surrogateescape")

            for match in re.findall(SCHEDLGU_REGEX_PATTERN, content):
                event = SchedLgU.from_line(match)

                yield SchedLgURecord(
                    ts=event.ts,
                    job=event.job,
                    command=event.command,
                    status=event.status,
                    exit_code=event.exit_code,
                    version=event.version,
                    _target=self.target,
                )



from flow.record.fieldtypes import datetime as dt

from dissect.target.filesystem import VirtualFilesystem
from dissect.target.plugins.os.unix.log.journal import JournalPlugin
from dissect.target.target import Target
from tests._utils import absolute_path


def test_journal_plugin(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    """test linux systemd journal file parsing."""

    data_file = absolute_path("_data/plugins/os/unix/log/journal/journal")
    fs_unix.map_file("var/log/journal/1337/user-1000.journal", data_file)
    target_unix.add_plugin(JournalPlugin)

    results = list(target_unix.journal())
    assert len(results) == 2

    record = results[0]
    assert record.ts == dt("2023-05-19T16:22:38.841870+00:00")
    assert record.message == (
        "Window manager warning: last_user_time (928062) is greater than comparison timestamp (928031).  "
        "This most likely represents a buggy client sending inaccurate timestamps in messages such as "
        "_NET_ACTIVE_WINDOW.  Trying to work around..."
    )
    assert record.syslog_facility == "3"
    assert record.syslog_identifier == "gnome-shell"
    assert record.pid == 2096
    assert record.transport == "stdout"
    assert record.source == "/var/log/journal/1337/user-1000.journal"



from __future__ import annotations

import io
import logging
from struct import error as StructError
from typing import BinaryIO, Iterator

from dissect.cstruct import cstruct
from dissect.ole import OLE
from dissect.ole.exceptions import Error as OleError
from dissect.shellitem.lnk import Lnk

from dissect.target import Target
from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.descriptor_extensions import UserRecordDescriptorExtension
from dissect.target.helpers.record import create_extended_descriptor
from dissect.target.helpers.shell_application_ids import APPLICATION_IDENTIFIERS
from dissect.target.helpers.utils import findall
from dissect.target.plugin import Plugin, export
from dissect.target.plugins.os.windows.lnk import LnkRecord, parse_lnk_file

log = logging.getLogger(__name__)

LNK_GUID = b"\x01\x14\x02\x00\x00\x00\x00\x00\xc0\x00\x00\x00\x00\x00\x00\x46"

JumpListRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
    "windows/jumplist",
    [
        ("string", "type"),
        ("string", "application_id"),
        ("string", "application_name"),
        *LnkRecord.target_fields,
    ],
)


custom_destination_def = """
struct header {
    int version;
    int unknown1;
    int unknown2;
    int value_type;
}

struct header_end {
    int number_of_entries;
}

struct header_end_0 {
    uint16  name_length;
    wchar   name[name_length];
    int     number_of_entries;
}

struct footer {
    char magic[4];
}
"""

c_custom_destination = cstruct()
c_custom_destination.load(custom_destination_def)


class JumpListFile:
    def __init__(self, fh: BinaryIO, file_name: str):
        self.fh = fh
        self.file_name = file_name

        self.application_id, self.application_type = file_name.split(".")
        self.application_type = self.application_type.split("-")[0]
        self.application_name = APPLICATION_IDENTIFIERS.get(self.application_id)

    def __iter__(self) -> Iterator[Lnk]:
        raise NotImplementedError

    @property
    def name(self) -> str:
        """Return the name of the application."""
        return self.application_name

    @property
    def id(self) -> str:
        """Return the application identifier."""
        return self.application_id

    @property
    def type(self) -> str:
        """Return the type of the Jump List file."""
        return self.application_type


class AutomaticDestinationFile(JumpListFile):
    """Parse Jump List AutomaticDestination file."""

    def __init__(self, fh: BinaryIO, file_name: str):
        super().__init__(fh, file_name)
        self.ole = OLE(self.fh)

    def __iter__(self) -> Iterator[Lnk]:
        for dir_name in self.ole.root.listdir():
            if dir_name == "DestList":
                continue

            dir = self.ole.get(dir_name)

            for item in dir.open():
                try:
                    yield Lnk(io.BytesIO(item))
                except StructError:
                    continue
                except Exception as e:
                    log.warning("Failed to parse LNK file from directory %s", dir_name)
                    log.debug("", exc_info=e)
                    continue


class CustomDestinationFile(JumpListFile):
    """Parse Jump List CustomDestination file."""

    MAGIC_FOOTER = 0xBABFFBAB
    VERSIONS = [2]

    def __init__(self, fh: BinaryIO, file_name: str):
        super().__init__(fh, file_name)

        self.fh.seek(-4, io.SEEK_END)
        self.footer = c_custom_destination.footer(self.fh.read(4))
        self.magic = int.from_bytes(self.footer.magic, "little")

        self.fh.seek(0, io.SEEK_SET)
        self.header = c_custom_destination.header(self.fh)
        self.version = self.header.version

        if self.header.value_type == 0:
            self.header_end = c_custom_destination.header_end_0(self.fh)
        elif self.header.value_type in [1, 2]:
            self.header_end = c_custom_destination.header_end(self.fh)
        else:
            raise NotImplementedError(
                f"The value_type ({self.header.value_type}) of the CustomDestination file is not implemented"
            )

        if self.version not in self.VERSIONS:
            raise NotImplementedError(f"The CustomDestination file has an unsupported version: {self.version}")

        if not self.MAGIC_FOOTER == self.magic:
            raise ValueError(f"The CustomDestination file has an invalid magic footer: {self.magic}")

    def __iter__(self) -> Iterator[Lnk]:
        # Searches for all LNK GUID's because the number of entries in the header is not always correct.
        buf = self.fh.read()

        for offset in findall(buf, LNK_GUID):
            try:
                lnk = Lnk(io.BytesIO(buf[offset + len(LNK_GUID) :]))
                yield lnk
            except EOFError:
                break
            except Exception as e:
                log.warning("Failed to parse LNK file from a CustomDestination file")
                log.debug("", exc_info=e)
                continue


class JumpListPlugin(Plugin):
    """Jump List is a Windows feature introduced in Windows 7.

    It stores information about recently accessed applications and files.

    References:
        - https://forensics.wiki/jump_lists
        - https://github.com/libyal/dtformats/blob/main/documentation/Jump%20lists%20format.asciidoc
    """

    __namespace__ = "jumplist"

    def __init__(self, target: Target):
        super().__init__(target)
        self.automatic_destinations = []
        self.custom_destinations = []

        for user_details in self.target.user_details.all_with_home():
            for destination in user_details.home_path.glob(
                "AppData/Roaming/Microsoft/Windows/Recent/CustomDestinations/*.customDestinations-ms"
            ):
                self.custom_destinations.append([destination, user_details.user])

        for user_details in self.target.user_details.all_with_home():
            for destination in user_details.home_path.glob(
                "AppData/Roaming/Microsoft/Windows/Recent/AutomaticDestinations/*.automaticDestinations-ms"
            ):
                self.automatic_destinations.append([destination, user_details.user])

    def check_compatible(self) -> None:
        if not any([self.automatic_destinations, self.custom_destinations]):
            raise UnsupportedPluginError("No Jump List files found")

    @export(record=JumpListRecord)
    def custom_destination(self) -> Iterator[JumpListRecord]:
        """Return the content of CustomDestination Windows Jump Lists.

        These are created when a user pins an application or a file in a Jump List.

        Yields JumpListRecord with fields:

        .. code-block:: text

            type (string): Type of Jump List.
            application_id (string): ID of the application.
            application_name (string): Name of the application.
            lnk_path (path): Path of the link (.lnk) file.
            lnk_name (string): Name of the link (.lnk) file.
            lnk_mtime (datetime): Modification time of the link (.lnk) file.
            lnk_atime (datetime): Access time of the link (.lnk) file.
            lnk_ctime (datetime): Creation time of the link (.lnk) file.
            lnk_relativepath (path): Relative path of target file to the link (.lnk) file.
            lnk_workdir (path): Path of the working directory the link (.lnk) file will execute from.
            lnk_iconlocation (path): Path of the display icon used for the link (.lnk) file.
            lnk_arguments (string): Command-line arguments passed to the target (linked) file.
            local_base_path (string): Absolute path of the target (linked) file.
            common_path_suffix (string): Suffix of the local_base_path.
            lnk_full_path (string): Full path of the linked file. Made from local_base_path and common_path_suffix.
            lnk_net_name (string): Specifies a server share path; for example, "\\\\server\\share".
            lnk_device_name (string): Specifies a device; for example, the drive letter "D:"
            machine_id (string): The NetBIOS name of the machine where the linked file was last known to reside.
            target_mtime (datetime): Modification time of the target (linked) file.
            target_atime (datetime): Access time of the target (linked) file.
            target_ctime (datetime): Creation time of the target (linked) file.
        """
        yield from self._generate_records(self.custom_destinations, CustomDestinationFile)

    @export(record=JumpListRecord)
    def automatic_destination(self) -> Iterator[JumpListRecord]:
        """Return the content of AutomaticDestination Windows Jump Lists.

        These are created automatically when a user opens an application or file.

        Yields JumpListRecord with fields:

        .. code-block:: text

            type (string): Type of Jump List.
            application_id (string): ID of the application.
            application_name (string): Name of the application.
            lnk_path (path): Path of the link (.lnk) file.
            lnk_name (string): Name of the link (.lnk) file.
            lnk_mtime (datetime): Modification time of the link (.lnk) file.
            lnk_atime (datetime): Access time of the link (.lnk) file.
            lnk_ctime (datetime): Creation time of the link (.lnk) file.
            lnk_relativepath (path): Relative path of target file to the link (.lnk) file.
            lnk_workdir (path): Path of the working directory the link (.lnk) file will execute from.
            lnk_iconlocation (path): Path of the display icon used for the link (.lnk) file.
            lnk_arguments (string): Command-line arguments passed to the target (linked) file.
            local_base_path (string): Absolute path of the target (linked) file.
            common_path_suffix (string): Suffix of the local_base_path.
            lnk_full_path (string): Full path of the linked file. Made from local_base_path and common_path_suffix.
            lnk_net_name (string): Specifies a server share path; for example, "\\\\server\\share".
            lnk_device_name (string): Specifies a device; for example, the drive letter "D:"
            machine_id (string): The NetBIOS name of the machine where the linked file was last known to reside.
            target_mtime (datetime): Modification time of the target (linked) file.
            target_atime (datetime): Access time of the target (linked) file.
            target_ctime (datetime): Creation time of the target (linked) file.
        """
        yield from self._generate_records(self.automatic_destinations, AutomaticDestinationFile)

    def _generate_records(
        self,
        destinations: list,
        destination_file: AutomaticDestinationFile | CustomDestinationFile,
    ) -> Iterator[JumpListRecord]:
        for destination, user in destinations:
            fh = destination.open("rb")

            try:
                jumplist = destination_file(fh, destination.name)
            except OleError:
                continue
            except Exception as e:
                self.target.log.warning("Failed to parse Jump List file: %s", destination)
                self.target.log.debug("", exc_info=e)
                continue

            for lnk in jumplist:
                if lnk := parse_lnk_file(self.target, lnk, destination):
                    yield JumpListRecord(
                        type=jumplist.type,
                        application_name=jumplist.name,
                        application_id=jumplist.id,
                        **lnk._asdict(),
                        _user=user,
                        _target=self.target,
                    )



import re
from datetime import datetime, timezone
from typing import Iterator

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.descriptor_extensions import UserRecordDescriptorExtension
from dissect.target.helpers.fsutil import TargetPath
from dissect.target.helpers.record import create_extended_descriptor
from dissect.target.plugin import export
from dissect.target.plugins.apps.remoteaccess.remoteaccess import (
    GENERIC_LOG_RECORD_FIELDS,
    RemoteAccessPlugin,
)
from dissect.target.plugins.general.users import UserDetails


class AnydeskPlugin(RemoteAccessPlugin):
    """Anydesk plugin."""

    __namespace__ = "anydesk"

    # Anydesk logs when installed as a service
    SERVICE_GLOBS = [
        # Standard client >= Windows 7
        "sysvol/ProgramData/AnyDesk/*.trace",
        # Custom client >= Windows 7
        "sysvol/ProgramData/AnyDesk/ad_*/*.trace",
        # Windows XP / 2003
        "sysvol/Documents and Settings/Public/AnyDesk/*.trace",
        "sysvol/Documents and Settings/Public/AnyDesk/ad_*/*.trace",
        # Standard/Custom client Linux/MacOS
        "var/log/anydesk*/*.trace",
    ]

    # User specific Anydesk logs
    USER_GLOBS = [
        # Standard client Windows
        "AppData/Roaming/AnyDesk/*.trace",
        # Custom client Windows
        "AppData/Roaming/AnyDesk/ad_*/*.trace",
        # Windows XP / 2003
        "AppData/AnyDesk/*.trace",
        # Standard client Linux/MacOS
        ".anydesk/*.trace",
        # Custom client Linux/MacOS
        ".anydesk_ad_*/*.trace",
    ]

    RemoteAccessLogRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
        "remoteaccess/anydesk/log", GENERIC_LOG_RECORD_FIELDS
    )

    def __init__(self, target):
        super().__init__(target)

        self.trace_files: set[tuple[TargetPath, UserDetails]] = set()

        # Service globs
        user = None
        for trace_glob in self.SERVICE_GLOBS:
            for trace_file in self.target.fs.path().glob(trace_glob):
                self.trace_files.add((trace_file, user))

        # User globs
        for user_details in self.target.user_details.all_with_home():
            for trace_glob in self.USER_GLOBS:
                for trace_file in user_details.home_path.glob(trace_glob):
                    self.trace_files.add((trace_file, user_details.user))

    def check_compatible(self) -> None:
        if not self.trace_files:
            raise UnsupportedPluginError("No Anydesk trace files found on target")

    @export(record=RemoteAccessLogRecord)
    def logs(self) -> Iterator[RemoteAccessLogRecord]:
        """Parse AnyDesk trace files.

        AnyDesk is a remote desktop application and can be used by adversaries to get (persistent) access to a machine.
        Log files (.trace files) can be stored on various locations, based on target OS and client type.
        Timestamps in trace files do not carry a time zone designator (TZD) but are in fact UTC.

        References:
            - https://www.inversecos.com/2021/02/forensic-analysis-of-anydesk-logs.html
            - https://support.anydesk.com/knowledge/trace-files#trace-file-locations
        """
        for trace_file, user in self.trace_files:
            for line in trace_file.open("rt", errors="backslashreplace"):
                line = line.strip()

                if not line or "* * * * * * * * * * * * * *" in line:
                    continue

                try:
                    level, ts_date, ts_time, message = line.split(" ", 3)

                    timestamp = datetime.strptime(f"{ts_date} {ts_time}", "%Y-%m-%d %H:%M:%S.%f").replace(
                        tzinfo=timezone.utc
                    )
                    message = re.sub(r"\s\s+", " ", f"{level} {message}")

                    yield self.RemoteAccessLogRecord(
                        ts=timestamp,
                        message=message,
                        source=trace_file,
                        _target=self.target,
                        _user=user,
                    )

                except ValueError as e:
                    self.target.log.warning("Could not parse log line in file %s: '%s'", trace_file, line)
                    self.target.log.debug("", exc_info=e)



import logging
import pathlib
from typing import BinaryIO, Iterator, Union

from dissect.fve import bde
from dissect.util.stream import AlignedStream

from dissect.target.exceptions import VolumeSystemError
from dissect.target.helpers.keychain import KeyType
from dissect.target.volume import EncryptedVolumeSystem, Volume

log = logging.getLogger(__name__)


class BitlockerVolumeSystemError(VolumeSystemError):
    pass


class BitlockerVolumeSystem(EncryptedVolumeSystem):
    __type__ = "bitlocker"

    def __init__(self, fh: Union[BinaryIO, list[BinaryIO]], *args, **kwargs):
        super().__init__(fh, *args, **kwargs)
        self.bde = bde.BDE(fh)

    @staticmethod
    def _detect(fh: BinaryIO) -> bool:
        return bde.is_bde_volume(fh)

    def _volumes(self) -> Iterator[Volume]:
        if isinstance(self.fh, Volume):
            volume_details = dict(
                number=self.fh.number,
                offset=self.fh.offset,
                vtype=self.fh.type,
                name=self.fh.name,
                guid=self.fh.guid,
            )
        else:
            volume_details = dict(
                number=1,
                offset=0,
                vtype=None,
                name=None,
            )

        stream = self.unlock_volume()
        yield Volume(
            fh=stream,
            size=stream.size,
            raw=self.fh,
            vs=self,
            **volume_details,
        )

    def unlock_with_passphrase(self, passphrase: str, is_wildcard: bool = False) -> None:
        try:
            self.bde.unlock_with_passphrase(passphrase)
            log.debug("Unlocked BDE volume with provided passphrase")
        except ValueError:
            if not is_wildcard:
                log.exception("Failed to unlock BDE volume with provided passphrase")

    def unlock_with_recovery_key(self, recovery_key: str, is_wildcard: bool = False) -> None:
        try:
            self.bde.unlock_with_recovery_password(recovery_key)
            log.debug("Unlocked BDE volume with recovery key")
        except ValueError:
            if not is_wildcard:
                log.exception("Failed to unlock BDE volume with recovery password")

    def unlock_with_bek_file(self, bek_file: pathlib.Path, is_wildcard: bool = False) -> None:
        if not bek_file.exists():
            if not is_wildcard:
                log.error("Provided BEK file does not exist: %s", bek_file)
            return

        with bek_file.open(mode="rb") as fh:
            try:
                self.bde.unlock_with_bek(fh)
                log.debug("Unlocked BDE volume with BEK file %s", bek_file)
            except ValueError:
                if not is_wildcard:
                    log.exception("Failed to unlock BDE volume with BEK file %s", bek_file)

    def unlock_volume(self) -> AlignedStream:
        if self.bde.has_clear_key():
            self.bde.unlock_with_clear_key()
        else:
            identifiers = [str(i) for i in self.bde.identifiers]
            keys = self.get_keys_for_identifiers(identifiers) + self.get_keys_without_identifier()

            for key in keys:
                if key.key_type == KeyType.PASSPHRASE and self.bde.has_passphrase():
                    self.unlock_with_passphrase(key.value, key.is_wildcard)
                elif key.key_type == KeyType.RECOVERY_KEY and self.bde.has_recovery_password():
                    self.unlock_with_recovery_key(key.value, key.is_wildcard)
                elif key.key_type == KeyType.FILE:
                    bek_file = pathlib.Path(key.value)
                    self.unlock_with_bek_file(bek_file, key.is_wildcard)

                if self.bde.unlocked:
                    log.info("Volume %s with identifiers %s unlocked with %s", self.fh, identifiers, key)
                    break

        if self.bde.unlocked:
            return self.bde.open()
        else:
            raise BitlockerVolumeSystemError("Failed to unlock BDE volume")



import io
from pathlib import Path
from typing import BinaryIO, Union

from dissect.hypervisor import vhdx

from dissect.target.container import Container


class VhdxContainer(Container):
    __type__ = "vhdx"

    def __init__(self, fh: Union[BinaryIO, Path], *args, **kwargs):
        self.vhdx = vhdx.VHDX(fh)
        super().__init__(fh, self.vhdx.size, *args, **kwargs)

    @staticmethod
    def _detect_fh(fh: BinaryIO, original: Union[list, BinaryIO]) -> bool:
        return fh.read(8) == b"vhdxfile"

    @staticmethod
    def detect_path(path: Path, original: Union[list, BinaryIO]) -> bool:
        return path.suffix.lower() in (".vhdx", ".avhdx")

    def read(self, length: int) -> bytes:
        return self.vhdx.read(length)

    def seek(self, offset: int, whence: int = io.SEEK_SET) -> int:
        return self.vhdx.seek(offset, whence)

    def tell(self) -> int:
        return self.vhdx.tell()

    def close(self) -> None:
        pass



from typing import Iterator, Optional

from dissect.shellitem.lnk import Lnk
from dissect.util import ts

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.fsutil import TargetPath
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, arg, export
from dissect.target.target import Target

LnkRecord = TargetRecordDescriptor(
    "windows/filesystem/lnk",
    [
        ("path", "lnk_path"),
        ("string", "lnk_name"),
        ("datetime", "lnk_mtime"),
        ("datetime", "lnk_atime"),
        ("datetime", "lnk_ctime"),
        ("path", "lnk_relativepath"),
        ("path", "lnk_workdir"),
        ("string", "lnk_arguments"),
        ("path", "lnk_iconlocation"),
        ("string", "local_base_path"),
        ("string", "common_path_suffix"),
        ("string", "lnk_net_name"),
        ("string", "lnk_device_name"),
        ("path", "lnk_full_path"),
        ("string", "machine_id"),
        ("datetime", "target_mtime"),
        ("datetime", "target_atime"),
        ("datetime", "target_ctime"),
    ],
)


def parse_lnk_file(target: Target, lnk_file: Lnk, lnk_path: TargetPath) -> Iterator[LnkRecord]:
    # we need to get the active codepage from the system to properly decode some values
    codepage = target.codepage or "ascii"

    lnk_net_name = lnk_device_name = None

    if lnk_file.link_header:
        lnk_name = lnk_file.stringdata.name_string.string if lnk_file.flag("has_name") else None

        lnk_mtime = ts.from_unix(lnk_path.stat().st_mtime)
        lnk_atime = ts.from_unix(lnk_path.stat().st_atime)
        lnk_ctime = ts.from_unix(lnk_path.stat().st_ctime)

        lnk_relativepath = (
            target.fs.path(lnk_file.stringdata.relative_path.string) if lnk_file.flag("has_relative_path") else None
        )
        lnk_workdir = (
            target.fs.path(lnk_file.stringdata.working_dir.string) if lnk_file.flag("has_working_dir") else None
        )
        lnk_iconlocation = (
            target.fs.path(lnk_file.stringdata.icon_location.string) if lnk_file.flag("has_icon_location") else None
        )
        lnk_arguments = lnk_file.stringdata.command_line_arguments.string if lnk_file.flag("has_arguments") else None
        local_base_path = (
            lnk_file.linkinfo.local_base_path.decode(codepage)
            if lnk_file.flag("has_link_info") and lnk_file.linkinfo.flag("volumeid_and_local_basepath")
            else None
        )
        common_path_suffix = (
            lnk_file.linkinfo.common_path_suffix.decode(codepage) if lnk_file.flag("has_link_info") else None
        )

        if local_base_path and common_path_suffix:
            lnk_full_path = target.fs.path(local_base_path + common_path_suffix)
        elif local_base_path and not common_path_suffix:
            lnk_full_path = target.fs.path(local_base_path)
        else:
            lnk_full_path = None

        if lnk_file.flag("has_link_info"):
            if lnk_file.linkinfo.flag("common_network_relative_link_and_pathsuffix"):
                lnk_net_name = (
                    lnk_file.linkinfo.common_network_relative_link.net_name.decode()
                    if lnk_file.linkinfo.common_network_relative_link.net_name
                    else None
                )
                lnk_device_name = (
                    lnk_file.linkinfo.common_network_relative_link.device_name.decode()
                    if lnk_file.linkinfo.common_network_relative_link.device_name
                    else None
                )
        try:
            machine_id = lnk_file.extradata.TRACKER_PROPS.machine_id.decode(codepage).strip("\x00")
        except AttributeError:
            machine_id = None

        target_mtime = ts.wintimestamp(lnk_file.link_header.write_time)
        target_atime = ts.wintimestamp(lnk_file.link_header.access_time)
        target_ctime = ts.wintimestamp(lnk_file.link_header.creation_time)

        return LnkRecord(
            lnk_path=lnk_path,
            lnk_name=lnk_name,
            lnk_mtime=lnk_mtime,
            lnk_atime=lnk_atime,
            lnk_ctime=lnk_ctime,
            lnk_relativepath=lnk_relativepath,
            lnk_workdir=lnk_workdir,
            lnk_iconlocation=lnk_iconlocation,
            lnk_arguments=lnk_arguments,
            local_base_path=local_base_path,
            common_path_suffix=common_path_suffix,
            lnk_full_path=lnk_full_path,
            lnk_net_name=lnk_net_name,
            lnk_device_name=lnk_device_name,
            machine_id=machine_id,
            target_mtime=target_mtime,
            target_atime=target_atime,
            target_ctime=target_ctime,
            _target=target,
        )


class LnkPlugin(Plugin):
    """Windows lnk plugin."""

    def __init__(self, target: Target) -> None:
        super().__init__(target)
        self.folders = ["programdata", "users", "windows"]

    def check_compatible(self) -> None:
        for folder in self.folders:
            if self.target.fs.path(f"sysvol/{folder}").exists():
                return None
        raise UnsupportedPluginError("No folders containing link files found")

    @arg("--path", "-p", dest="path", default=None, help="Path to directory or .lnk file in target")
    @export(record=LnkRecord)
    def lnk(self, path: Optional[str] = None) -> Iterator[LnkRecord]:
        """Parse all .lnk files in /ProgramData, /Users, and /Windows or from a specified path in record format.

        Yields a LnkRecord record with the following fields:

        .. code-block:: text

            lnk_path (path): Path of the link (.lnk) file.
            lnk_name (string): Name of the link (.lnk) file.
            lnk_mtime (datetime): Modification time of the link (.lnk) file.
            lnk_atime (datetime): Access time of the link (.lnk) file.
            lnk_ctime (datetime): Creation time of the link (.lnk) file.
            lnk_relativepath (path): Relative path of target file to the link (.lnk) file.
            lnk_workdir (path): Path of the working directory the link (.lnk) file will execute from.
            lnk_iconlocation (path): Path of the display icon used for the link (.lnk) file.
            lnk_arguments (string): Command-line arguments passed to the target (linked) file.
            local_base_path (string): Absolute path of the target (linked) file.
            common_path_suffix (string): Suffix of the local_base_path.
            lnk_full_path (string): Full path of the linked file. Made from local_base_path and common_path_suffix.
            lnk_net_name (string): Specifies a server share path; for example, "\\\\server\\share".
            lnk_device_name (string): Specifies a device; for example, the drive letter "D:"
            machine_id (string): The NetBIOS name of the machine where the linked file was last known to reside.
            target_mtime (datetime): Modification time of the target (linked) file.
            target_atime (datetime): Access time of the target (linked) file.
            target_ctime (datetime): Creation time of the target (linked) file.
        """

        for entry in self.lnk_entries(path):
            lnk_file = Lnk(entry.open())
            yield parse_lnk_file(self.target, lnk_file, entry)

    def lnk_entries(self, path: Optional[str] = None) -> Iterator[TargetPath]:
        if path:
            target_path = self.target.fs.path(path)
            if not target_path.exists():
                self.target.log.error("Provided path %s does not exist on target", target_path)
                return

            if target_path.is_file():
                yield target_path
            else:
                yield from target_path.rglob("*.lnk")
        else:
            for folder in self.folders:
                yield from self.target.fs.path("sysvol").joinpath(folder).rglob("*.lnk")



import csv
import gzip
from typing import Iterator

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, export

AcquireHashRecord = TargetRecordDescriptor(
    "filesystem/acquire_hash",
    [
        ("path", "path"),
        ("filesize", "filesize"),
        ("digest", "digest"),
    ],
)


class AcquireHashPlugin(Plugin):
    """Plugin to return file hashes collected by Acquire."""

    def __init__(self, target):
        super().__init__(target)
        self.hash_file = target.fs.path("$metadata$/file-hashes.csv.gz")

    def check_compatible(self) -> None:
        if not self.hash_file.exists():
            raise UnsupportedPluginError("No hash file found")

    @export(record=AcquireHashRecord)
    def acquire_hashes(self) -> Iterator[AcquireHashRecord]:
        """Return file hashes collected by Acquire.

        An Acquire file container contains a file hashes csv when the hashes module was used. The content of this csv
        file is returned.
        """

        with self.hash_file.open() as fh:
            for row in csv.DictReader(gzip.open(fh, "rt")):
                yield AcquireHashRecord(
                    path=self.target.fs.path((row["path"])),
                    filesize=row["file-size"],
                    digest=(row["md5"] or None, row["sha1"] or None, row["sha256"] or None),
                    _target=self.target,
                )



from unittest.mock import call, patch

from dissect.target.loaders.pvm import PvmLoader
from tests._utils import mkdirs


@patch("dissect.target.loaders.pvs.HddContainer")
@patch("dissect.target.loaders.pvs.pvs.PVS")
def test_pvm_loader(PVS, HddContainer, target_bare, tmp_path):
    mkdirs(tmp_path, ["Test.pvm"])
    (tmp_path / "Test.pvm" / "config.pvs").touch()

    PVS.return_value = PVS
    PVS.disks.return_value = ["mock.hdd"]
    HddContainer.return_value = HddContainer

    pvm_loader = PvmLoader(tmp_path / "Test.pvm")
    pvm_loader.map(target_bare)

    assert len(target_bare.disks) == 1
    assert HddContainer.mock_calls == [call(tmp_path.resolve() / "Test.pvm" / "mock.hdd")]



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

from dissect.target.loaders.dir import DirLoader, find_and_map_dirs, find_dirs
from dissect.target.plugin import OperatingSystem

if TYPE_CHECKING:
    from dissect.target import Target


# KAPE doesn't have the correct filenames for several files, like $J or $Secure_$SDS
# The same applies to Velociraptor offline collector Windows.KapeFiles.Targets
USNJRNL_PATHS = ["$Extend/$J", "$Extend/$UsnJrnl$J"]


class KapeLoader(DirLoader):
    """Load KAPE forensic image format files.

    References:
        - https://www.kroll.com/en/insights/publications/cyber/kroll-artifact-parser-extractor-kape
    """

    @staticmethod
    def detect(path: Path) -> bool:
        os_type, dirs = find_dirs(path)
        if os_type == OperatingSystem.WINDOWS:
            for dir_path in dirs:
                for path in USNJRNL_PATHS:
                    if dir_path.joinpath(path).exists():
                        return True

        return False

    def map(self, target: Target) -> None:
        find_and_map_dirs(
            target,
            self.path,
            sds_path="$Secure_$SDS",
            usnjrnl_path=USNJRNL_PATHS,
        )



from dissect.target.filesystem import VirtualFilesystem
from dissect.target.plugins.os.unix.linux.proc import ProcPlugin
from dissect.target.target import Target


def test_environ(target_linux_users: Target, fs_linux_proc: VirtualFilesystem):
    target_linux_users.add_plugin(ProcPlugin)
    results = list(target_linux_users.environ())
    assert len(results) == 4



from typing import Iterator

import pytest

from dissect.target import Target
from dissect.target.filesystem import VirtualFilesystem
from dissect.target.plugins.apps.browser import brave
from tests._utils import absolute_path


@pytest.fixture
def target_brave(target_win_users: Target, fs_win: VirtualFilesystem) -> Iterator[Target]:
    base_path = "Users\\John\\AppData\\Local\\BraveSoftware\\Brave-Browser\\User Data\\Default"
    files = [
        ("History", "_data/plugins/apps/browser/chrome/History"),
        ("Preferences", "_data/plugins/apps/browser/chrome/Preferences"),
        ("Secure Preferences", "_data/plugins/apps/browser/chrome/Secure Preferences"),
        ("Network\\Cookies", "_data/plugins/apps/browser/chromium/Cookies"),
        ("Login Data", "_data/plugins/apps/browser/chromium/Login Data"),
    ]

    for filename, test_path in files:
        fs_win.map_file("\\".join([base_path, filename]), absolute_path(test_path))

    target_win_users.add_plugin(brave.BravePlugin)

    yield target_win_users


def test_brave_history(target_brave: Target) -> None:
    records = list(target_brave.brave.history())
    assert len(records) == 5


def test_brave_downloads(target_brave: Target) -> None:
    records = list(target_brave.brave.downloads())
    assert len(records) == 1


def test_brave_extensions(target_brave: Target) -> None:
    records = list(target_brave.brave.extensions())
    assert len(records) == 8


def test_brave_cookies(target_brave: Target) -> None:
    records = list(target_brave.brave.cookies())
    assert len(records) == 5
    assert all(record.host == ".tweakers.net" for record in records)


def test_windows_edge_passwords_plugin(target_brave: Target) -> None:
    records = list(target_brave.brave.passwords())

    assert len(records) == 2

    for record in records:
        assert record.browser == "brave"
        assert record.decrypted_username == "username"
        assert record.decrypted_password is None

    assert records[0].url == "https://example.com/"
    assert records[1].url == "https://example.org/"



from dissect.target.loaders.dir import DirLoader, find_dirs
from dissect.target.plugin import OperatingSystem
from tests._utils import mkdirs


def test_dir_loader_windows(target_bare, tmp_path):
    root = tmp_path
    mkdirs(root, ["windows/system32"])

    os_type, dirs = find_dirs(root)
    assert os_type == OperatingSystem.WINDOWS
    assert len(dirs) == 1

    assert DirLoader.detect(root)

    loader = DirLoader(root)
    loader.map(target_bare)

    assert len(target_bare.filesystems) == 1


def test_dir_loader_winnt(target_bare, tmp_path):
    root = tmp_path
    mkdirs(tmp_path, ["winnt"])

    os_type, dirs = find_dirs(root)
    assert os_type == OperatingSystem.WINDOWS
    assert len(dirs) == 1

    assert DirLoader.detect(root)

    loader = DirLoader(root)
    loader.map(target_bare)

    assert len(target_bare.filesystems) == 1


def test_dir_loader_linux(target_bare, tmp_path):
    root = tmp_path
    mkdirs(root, ["etc", "var"])

    os_type, dirs = find_dirs(root)
    assert os_type == OperatingSystem.LINUX
    assert len(dirs) == 1

    assert DirLoader.detect(root)

    loader = DirLoader(root)
    loader.map(target_bare)

    assert len(target_bare.filesystems) == 1


def test_dir_loader_macos(target_bare, tmp_path):
    root = tmp_path
    mkdirs(root, ["Library"])

    os_type, dirs = find_dirs(root)
    assert os_type == OperatingSystem.OSX
    assert len(dirs) == 1

    assert DirLoader.detect(root)

    loader = DirLoader(root)
    loader.map(target_bare)

    assert len(target_bare.filesystems) == 1


def test_dir_loader_windows_drive_letters(target_bare, tmp_path):
    root = tmp_path
    mkdirs(root, ["C/windows/system32", "D/test", "E/test"])

    os_type, dirs = find_dirs(root)
    assert os_type == OperatingSystem.WINDOWS
    assert len(dirs) == 3

    assert DirLoader.detect(root)

    loader = DirLoader(root)
    loader.map(target_bare)

    assert len(target_bare.filesystems) == 3
    assert len(target_bare.fs.mounts) == 3



from __future__ import annotations

import os
import stat
from datetime import datetime, timezone
from typing import TextIO

from dissect.target.exceptions import FileNotFoundError
from dissect.target.filesystem import FilesystemEntry, LayerFilesystemEntry
from dissect.target.helpers import fsutil
from dissect.target.helpers.fsutil import TargetPath

# ['mode', 'addr', 'dev', 'nlink', 'uid', 'gid', 'size', 'atime', 'mtime', 'ctime']
STAT_TEMPLATE = """  File: {path} {symlink}
  Size: {size}       Blocks: {blocks}    IO Block: {blksize}     {filetype}
Device: {device}     Inode: {inode}      Links: {nlink}
Access: ({modeord}/{modestr})  Uid: ( {uid} )   Gid: ( {gid} )
Access: {atime}
Modify: {mtime}
Change: {ctime}
 Birth: {btime}"""

FALLBACK_LS_COLORS = "rs=0:di=01;34:ln=01;36:mh=00:pi=40;33:so=01;35:do=01;35:bd=40;33;01:cd=40;33;01:or=40;31;01:mi=00:su=37;41:sg=30;43:ca=30;41:tw=30;42:ow=34;42:st=37;44:ex=01;32"  # noqa: E501


def prepare_ls_colors() -> dict[str, str]:
    """Parse the LS_COLORS environment variable so we can use it later."""
    d = {}
    ls_colors = os.environ.get("LS_COLORS", FALLBACK_LS_COLORS)
    for line in ls_colors.split(":"):
        if not line:
            continue

        ft, _, value = line.partition("=")
        if ft.startswith("*"):
            ft = ft[1:]

        d[ft] = f"\x1b[{value}m{{}}\x1b[0m"

    return d


LS_COLORS = prepare_ls_colors()


def fmt_ls_colors(ft: str, name: str) -> str:
    """Helper method to colorize strings according to LS_COLORS."""
    try:
        return LS_COLORS[ft].format(name)
    except KeyError:
        pass

    try:
        return LS_COLORS[fsutil.splitext(name)[1]].format(name)
    except KeyError:
        pass

    return name


def human_size(bytes: int, units: list[str] = ["", "K", "M", "G", "T", "P", "E"]) -> str:
    """Helper function to return the human readable string representation of bytes."""
    return str(bytes) + units[0] if bytes < 1024 else human_size(bytes >> 10, units[1:])


def stat_modestr(st: fsutil.stat_result) -> str:
    """Helper method for generating a mode string from a numerical mode value."""
    return stat.filemode(st.st_mode)


def print_extensive_file_stat_listing(
    stdout: TextIO,
    name: str,
    entry: FilesystemEntry | None = None,
    timestamp: datetime | None = None,
    human_readable: bool = False,
) -> None:
    """Print the file status as a single line."""
    if entry is not None:
        try:
            entry_stat = entry.lstat()
            if timestamp is None:
                timestamp = entry_stat.st_mtime
            symlink = f" -> {entry.readlink()}" if entry.is_symlink() else ""
            utc_time = datetime.fromtimestamp(timestamp, tz=timezone.utc).isoformat(timespec="microseconds")
            size = f"{human_size(entry_stat.st_size):5s}" if human_readable else f"{entry_stat.st_size:10d}"

            print(
                (
                    f"{stat_modestr(entry_stat)} {entry_stat.st_uid:4d} {entry_stat.st_gid:4d} {size} "
                    f"{utc_time} {name}{symlink}"
                ),
                file=stdout,
            )
            return
        except FileNotFoundError:
            pass

    hr_spaces = f"{'':5s}" if human_readable else " "
    regular_spaces = f"{'':10s}" if not human_readable else " "

    print(f"??????????    ?    ?{regular_spaces}?{hr_spaces}????-??-??T??:??:??.??????+??:?? {name}", file=stdout)


def ls_scandir(path: fsutil.TargetPath, color: bool = False) -> list[tuple[fsutil.TargetPath, str]]:
    """List a directory for the given path."""
    result = []
    if not path.exists() or not path.is_dir():
        return []

    for file_ in path.iterdir():
        file_type = None
        if color:
            if file_.is_symlink():
                file_type = "ln"
            elif file_.is_dir():
                file_type = "di"
            elif file_.is_file():
                file_type = "fi"

        result.append((file_, fmt_ls_colors(file_type, file_.name) if color else file_.name))

        # If we happen to scan an NTFS filesystem see if any of the
        # entries has an alternative data stream and also list them.
        entry = file_.get()
        if isinstance(entry, LayerFilesystemEntry):
            if entry.entries.fs.__type__ == "ntfs":
                attrs = entry.lattr()
                for data_stream in attrs.DATA:
                    if data_stream.name != "":
                        name = f"{file_.name}:{data_stream.name}"
                        result.append((file_, fmt_ls_colors(file_type, name) if color else name))

    result.sort(key=lambda e: e[0].name)

    return result


def print_ls(
    path: fsutil.TargetPath,
    depth: int,
    stdout: TextIO,
    long_listing: bool = False,
    human_readable: bool = False,
    recursive: bool = False,
    use_ctime: bool = False,
    use_atime: bool = False,
    color: bool = True,
) -> None:
    """Print ls output"""
    subdirs = []

    if path.is_dir():
        contents = ls_scandir(path, color)
    elif path.is_file():
        contents = [(path, path.name)]

    if depth > 0:
        print(f"\n{str(path)}:", file=stdout)

    if not long_listing:
        for target_path, name in contents:
            print(name, file=stdout)
            if target_path.is_dir():
                subdirs.append(target_path)
    else:
        if len(contents) > 1:
            print(f"total {len(contents)}", file=stdout)
        for target_path, name in contents:
            try:
                entry = target_path.get()
                entry_stat = entry.lstat()
                show_time = entry_stat.st_mtime
                if use_ctime:
                    show_time = entry_stat.st_ctime
                elif use_atime:
                    show_time = entry_stat.st_atime
            except FileNotFoundError:
                entry = None
                show_time = None
            print_extensive_file_stat_listing(stdout, name, entry, show_time, human_readable)
            if target_path.is_dir():
                subdirs.append(target_path)

    if recursive and subdirs:
        for subdir in subdirs:
            print_ls(subdir, depth + 1, stdout, long_listing, human_readable, recursive, use_ctime, use_atime, color)


def print_stat(path: fsutil.TargetPath, stdout: TextIO, dereference: bool = False) -> None:
    """Print file status."""
    symlink = f"-> {path.readlink()}" if path.is_symlink() else ""
    s = path.stat() if dereference else path.lstat()

    def filetype(path: TargetPath) -> str:
        if path.is_dir():
            return "directory"
        elif path.is_symlink():
            return "symbolic link"
        elif path.is_file():
            return "regular file"

    res = STAT_TEMPLATE.format(
        path=path,
        symlink=symlink,
        size=s.st_size,
        filetype=filetype(path),
        device="?",
        inode=s.st_ino,
        blocks=s.st_blocks if s.st_blocks is not None else "?",
        blksize=s.st_blksize or "?",
        nlink=s.st_nlink,
        modeord=oct(stat.S_IMODE(s.st_mode)),
        modestr=stat_modestr(s),
        uid=s.st_uid,
        gid=s.st_gid,
        atime=datetime.fromtimestamp(s.st_atime, tz=timezone.utc).isoformat(timespec="microseconds"),
        mtime=datetime.fromtimestamp(s.st_mtime, tz=timezone.utc).isoformat(timespec="microseconds"),
        ctime=datetime.fromtimestamp(s.st_ctime, tz=timezone.utc).isoformat(timespec="microseconds"),
        btime=datetime.fromtimestamp(s.st_birthtime, tz=timezone.utc).isoformat(timespec="microseconds")
        if hasattr(s, "st_birthtime") and s.st_birthtime
        else "?",
    )
    print(res, file=stdout)

    try:
        if (xattr := path.get().attr()) and isinstance(xattr, list) and hasattr(xattr[0], "name"):
            print("  Attr:")
            print_xattr(path.name, xattr, stdout)
    except Exception:
        pass


def print_xattr(basename: str, xattr: list, stdout: TextIO) -> None:
    """Mimics getfattr -d {file} behaviour."""
    if not hasattr(xattr[0], "name"):
        return

    XATTR_TEMPLATE = "# file: {basename}\n{attrs}"
    res = XATTR_TEMPLATE.format(
        basename=basename, attrs="\n".join([f'{attr.name}="{attr.value.decode()}"' for attr in xattr])
    )
    print(res, file=stdout)



from collections import defaultdict
from typing import BinaryIO, Iterator, Union

from dissect.volume.md.md import MD, MDPhysicalDisk, find_super_block

from dissect.target.volume import LogicalVolumeSystem, Volume


class MdVolumeSystem(LogicalVolumeSystem):
    __type__ = "md"

    def __init__(self, fh: Union[BinaryIO, list[BinaryIO]], *args, **kwargs):
        self.md = MD(fh)
        super().__init__(fh, *args, **kwargs)

    @classmethod
    def open_all(cls, volumes: list[BinaryIO]) -> Iterator[LogicalVolumeSystem]:
        devices = defaultdict(list)

        for vol in volumes:
            if not cls.detect_volume(vol):
                continue

            device = MDPhysicalDisk(vol)
            devices[device.set_uuid].append(device)

        for devs in devices.values():
            try:
                yield cls(devs)
            except Exception:
                continue

    @staticmethod
    def _detect(fh: BinaryIO) -> bool:
        vols = [fh] if not isinstance(fh, list) else fh
        for vol in vols:
            if MdVolumeSystem.detect_volume(vol):
                return True
        return False

    @staticmethod
    def _detect_volume(fh: BinaryIO) -> bool:
        offset, _, _ = find_super_block(fh)
        return offset is not None

    def _volumes(self) -> Iterator[Volume]:
        # MD only supports one configuration and virtual disk but doing this as a loop
        # makes it automatically safe for empty configurations
        for conf in self.md.configurations:
            for vd in conf.virtual_disks:
                fh = vd.open()
                yield Volume(fh, 1, None, vd.size, None, vd.name, vd.uuid, raw=self.md, vs=self)



import pytest

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.filesystem import VirtualFilesystem
from dissect.target.plugins.os.unix.linux.proc import ProcPlugin, ProcProcess
from dissect.target.target import Target


def test_process(target_linux_users: Target, fs_linux_proc: VirtualFilesystem):
    target_linux_users.add_plugin(ProcPlugin)

    process = target_linux_users.proc.process(1)
    assert isinstance(process, ProcProcess)
    assert process.pid == 1
    assert process.name == "systemd"
    assert process.parent.name == "swapper"
    assert process.ppid == 0
    assert process.state == "Sleeping"
    assert str(process.runtime) == "1 day, 13:19:27.970000"
    assert process.starttime.isoformat() == "2023-04-03T22:10:54.300000+00:00"

    environ = list(process.environ())
    assert environ[0].variable == "VAR"
    assert environ[0].contents == "1"


def test_process_not_found(target_linux_users: Target, fs_linux_proc: VirtualFilesystem):
    target_linux_users.add_plugin(ProcPlugin)
    with pytest.raises(ProcessLookupError) as exc:
        target_linux_users.proc.process(404)
    assert str(exc.value) == f"Process with PID 404 could not be found on target: {target_linux_users}"


def test_processes(target_linux_users: Target, fs_linux_proc: VirtualFilesystem):
    target_linux_users.add_plugin(ProcPlugin)

    for process in target_linux_users.proc.processes():
        assert process.pid in (1, 2, 3, 1337)
        assert process.state in ("Sleeping", "Waking", "Running", "Wakekill")
        assert process.name in ("systemd", "kthread", "acquire", "sshd")
        assert process.starttime.isoformat() == "2023-04-03T22:10:54.300000+00:00"
        assert str(process.runtime) == "1 day, 13:19:27.970000"

        for env in process.environ():
            assert env.variable == "VAR"
            assert env.contents == "1"


def test_proc_plugin_incompatible(target_linux_users: Target, fs_linux: VirtualFilesystem):
    with pytest.raises(UnsupportedPluginError, match="No /proc directory found"):
        target_linux_users.add_plugin(ProcPlugin)



import re
from datetime import datetime, timezone
from typing import Iterator

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.descriptor_extensions import UserRecordDescriptorExtension
from dissect.target.helpers.fsutil import TargetPath
from dissect.target.helpers.record import create_extended_descriptor
from dissect.target.plugin import export
from dissect.target.plugins.apps.remoteaccess.remoteaccess import (
    GENERIC_LOG_RECORD_FIELDS,
    RemoteAccessPlugin,
)
from dissect.target.plugins.general.users import UserDetails

START_PATTERN = re.compile(r"^(\d{2}|\d{4})/")


class TeamViewerPlugin(RemoteAccessPlugin):
    """TeamViewer client plugin.

    Resources:
        - https://teamviewer.com/en/global/support/knowledge-base/teamviewer-classic/contact-support/find-your-log-files
        - https://www.systoolsgroup.com/forensics/teamviewer/
        - https://benleeyr.wordpress.com/2020/05/19/teamviewer-forensics-tested-on-v15/
    """

    __namespace__ = "teamviewer"

    SYSTEM_GLOBS = [
        "sysvol/Program Files/TeamViewer/*.log",
        "sysvol/Program Files (x86)/TeamViewer/*.log",
        "/var/log/teamviewer*/*.log",
    ]

    USER_GLOBS = [
        "AppData/Roaming/TeamViewer/teamviewer*_logfile.log",
        "Library/Logs/TeamViewer/teamviewer*_logfile*.log",
    ]

    RemoteAccessLogRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
        "remoteaccess/teamviewer/log", GENERIC_LOG_RECORD_FIELDS
    )

    def __init__(self, target):
        super().__init__(target)

        self.logfiles: list[list[TargetPath, UserDetails]] = []

        # Find system service log files.
        for log_glob in self.SYSTEM_GLOBS:
            for logfile in self.target.fs.glob(log_glob):
                self.logfiles.append([logfile, None])

        # Find user log files.
        for user_details in self.target.user_details.all_with_home():
            for log_glob in self.USER_GLOBS:
                for logfile in user_details.home_path.glob(log_glob):
                    self.logfiles.append([logfile, user_details])

    def check_compatible(self) -> None:
        if not len(self.logfiles):
            raise UnsupportedPluginError("No Teamviewer logs found")

    @export(record=RemoteAccessLogRecord)
    def logs(self) -> Iterator[RemoteAccessLogRecord]:
        """Yield TeamViewer client logs.

        TeamViewer is a commercial remote desktop application. An adversary may use it to gain persistence on a system.
        """
        for logfile, user_details in self.logfiles:
            logfile = self.target.fs.path(logfile)

            start_date = None
            with logfile.open("rt") as file:
                while True:
                    try:
                        line = file.readline()
                    except UnicodeDecodeError:
                        continue

                    # End of file, quit while loop
                    if not line:
                        break

                    line = line.strip()

                    # Skip empty lines
                    if not line:
                        continue
                    # Older logs first mention the start time and then leave out the year
                    if line.startswith("Start:"):
                        start_date = datetime.strptime(line.split()[1], "%Y/%m/%d")

                    # Sometimes there are weird, mult-line/pretty print log messages.
                    # We only parse the start line which starts with year (%Y/) or month (%m/)
                    if not re.match(START_PATTERN, line):
                        continue

                    ts_day, ts_time, message = line.split(" ", 2)
                    ts_time = ts_time.split(".")[0]

                    # Correct for use of : as millisecond separator
                    if ts_time.count(":") > 2:
                        ts_time = ":".join(ts_time.split(":")[:3])
                    # Correct for missing year in date
                    if ts_day.count("/") == 1:
                        if not start_date:
                            self.target.log.debug("Missing year in log line, skipping line.")
                            continue
                        ts_day = f"{start_date.year}/{ts_day}"
                    # Correct for year if short notation for 2000 is used
                    if ts_day.count("/") == 2 and len(ts_day.split("/")[0]) == 2:
                        ts_day = "20" + ts_day

                    timestamp = datetime.strptime(f"{ts_day} {ts_time}", "%Y/%m/%d %H:%M:%S").replace(
                        tzinfo=timezone.utc
                    )

                    yield self.RemoteAccessLogRecord(
                        ts=timestamp,
                        message=message,
                        source=logfile,
                        _target=self.target,
                        _user=user_details.user if user_details else None,
                    )



import io
from typing import BinaryIO, Callable, Dict, Generator, List, Tuple

from dissect.target import plugin
from dissect.target.helpers.record import TargetRecordDescriptor


class ScrapePlugin(plugin.Plugin):
    __namespace__ = "scrape"

    def check_compatible(self) -> None:
        pass

    @plugin.internal
    def find_needles(
        self,
        fh: BinaryIO,
        needles: List[bytes],
        block_size: int = io.DEFAULT_BUFFER_SIZE,
    ) -> Generator[Tuple[bytes, int], None, None]:
        """Yields needles and their offsets found in provided byte stream"""

        if not needles:
            raise ValueError("At least one needle value must be provided")

        max_needle_len = max([len(n) for n in needles])
        overlap_len = max_needle_len
        current_block = b""

        while True:
            next_block_offset = fh.tell()
            next_block = fh.read(block_size)

            if not next_block:
                break

            overlap = current_block[-overlap_len:] if overlap_len > 0 else b""
            current_block = overlap + next_block

            current_block_offset = next_block_offset - len(overlap)

            # Look for a needle in a current block
            last_needle_pos = -1
            last_needle_end = 0
            while True:
                needle_positions = [
                    (pos, needle)
                    for pos, needle in [(current_block.find(needle, last_needle_pos + 1), needle) for needle in needles]
                    if pos > -1
                ]

                if not needle_positions:
                    break

                closest_needle_pos, closest_needle = min(needle_positions)

                # just in case `fh` is changed outside the function
                saved_offset = fh.tell()
                try:
                    yield closest_needle, current_block_offset + closest_needle_pos
                finally:
                    fh.seek(saved_offset)

                last_needle_pos = closest_needle_pos
                last_needle_end = last_needle_pos + 1

            # the size of the data from the current block that will be prepended to the next block
            overlap_len = min(len(current_block) - last_needle_end, max_needle_len)

    @plugin.internal
    def find_needle_chunks(
        self,
        fh: BinaryIO,
        needle_chunk_size_map: Dict[bytes, int],
        block_size: int = io.DEFAULT_BUFFER_SIZE,
        chunk_reader: Callable[[BinaryIO, bytes, int, int], bytes] = None,
    ) -> Generator[Tuple[bytes, int, bytes], None, None]:
        """Yields tuples with an offset, a needle and a byte chunk found in provided byte stream."""

        chunk_reader = chunk_reader or read_plain_chunk

        needles = list(needle_chunk_size_map.keys())
        for needle, offset in self.find_needles(fh, needles=needles, block_size=block_size):
            saved_offset = fh.tell()
            chunk_size = needle_chunk_size_map[needle]
            try:
                yield (needle, offset, chunk_reader(fh, needle, offset, chunk_size))
            finally:
                fh.seek(saved_offset)

    @plugin.internal
    def scrape_chunks(
        self,
        fh: BinaryIO,
        needle_chunk_size_map: Dict[bytes, int],
        chunk_parser: Callable[[bytes, bytes], Generator],
        chunk_reader: Callable[[BinaryIO, bytes, int, int], bytes] = None,
        block_size: int = io.DEFAULT_BUFFER_SIZE,
    ) -> Generator[TargetRecordDescriptor, None, None]:
        """Yields records scraped from chunks found in a provided byte stream"""

        chunk_count = 0
        record_count = 0

        chunk_stream = self.find_needle_chunks(
            fh,
            needle_chunk_size_map,
            block_size=block_size,
            chunk_reader=chunk_reader,
        )

        for needle, _, chunk in chunk_stream:
            try:
                for record in chunk_parser(needle, chunk):
                    yield record
                    record_count += 1
                chunk_count += 1
            except Exception as e:
                self.target.log.warn(f"Chunk parsing failed with {e!r}")

        self.target.log.info(f"Scraped {chunk_count} chunks with {record_count} records")

    @plugin.internal
    def scrape_chunks_from_disks(
        self,
        chunk_parser: Callable[[bytes, bytes], Generator],
        needle_chunk_size_map: Dict[bytes, int] = None,
        needle: bytes = None,
        chunk_size: int = None,
        chunk_reader: Callable[[BinaryIO, bytes, int, int], bytes] = None,
        block_size: int = io.DEFAULT_BUFFER_SIZE,
    ) -> Generator[TargetRecordDescriptor, None, None]:
        """Yields records scraped from chunks found in target.disks"""

        if needle_chunk_size_map and (needle or chunk_size):
            raise ValueError("Either `needle_chunk_size_map` or both `needle` and `chunk_size` must be provided")

        if (needle and not chunk_size) or (not needle and chunk_size):
            raise ValueError("Both `needle` and `chunk_size` must be provided")
        else:
            needle_chunk_size_map = {needle: chunk_size}

        for disk in self.target.disks:
            # reset pointer to the start of a disk
            disk.seek(0)
            yield from self.scrape_chunks(
                disk,
                needle_chunk_size_map=needle_chunk_size_map,
                chunk_parser=chunk_parser,
                chunk_reader=chunk_reader,
                block_size=block_size,
            )

    @plugin.internal
    def scrape_needles_from_disks(
        self,
        needles: List[bytes] = None,
        needle: bytes = None,
        block_size: int = io.DEFAULT_BUFFER_SIZE,
    ) -> Generator[Tuple[BinaryIO, bytes, int], None, None]:
        """Yields bytestream / needle / offset tuples, scraped from target.disks"""

        if needles and needle:
            raise ValueError("Either `needles` values or a signle `needle` value must be provided")
        elif not needles and not needle:
            raise ValueError("At least one needle value must be provided")

        if needle:
            needles = [needle]

        for disk in self.target.disks:
            # reset pointer to the start of a disk
            disk.seek(0)
            for needle, offset in self.find_needles(disk, needles=needles, block_size=block_size):
                yield disk, needle, offset


def read_plain_chunk(fh: BinaryIO, needle: bytes, offset: int, chunk_size: int) -> bytes:
    """Read chunk from a byte stream for provided needle, offset and chunk size"""
    fh.seek(offset)
    return fh.read(chunk_size)



from io import BytesIO
from unittest import mock

import pytest

from dissect.target.filesystem import VirtualFilesystem
from dissect.target.loaders.hyperv import HyperVLoader
from tests._utils import absolute_path


@pytest.mark.parametrize(
    "descriptor_dir, disk_dir",
    [
        # Disk next to descriptor file
        ("c:", "c:"),
        # Disk in "Virtual Hard Disks" directory, relative from descriptor file
        ("c:", "c:\\Virtual Hard Disks"),
        # Disk in "Virtual Hard Disks" directory, one directory up relative from descriptor file
        ("c:\\Virtual Machines", "c:\\Virtual Hard Disks"),
        # Disk in absolute location
        ("c:\\Virtual Machines", "c:\\Disks"),
    ],
)
def test_hyperv_loader_xml(descriptor_dir, disk_dir, target_bare):
    gen1_xml_filename = "B90AC31B-C6F8-479F-9B91-07B894A6A3F6.xml"
    gen2_xml_filename = "D351C151-DAC7-4042-B434-B72D522C1E4A.xml"
    gen1_vmcx_filename = "EC04F346-DB96-4700-AF5B-77B3C56C38BD.vmcx"
    gen2_vmcx_filename = "993F7B33-6057-4D1E-A1FE-A1A1D77BE974.vmcx"

    vfs = VirtualFilesystem(case_sensitive=False, alt_separator="\\")
    vfs.map_file(f"{descriptor_dir}\\{gen1_xml_filename}", absolute_path(f"_data/loaders/hyperv/{gen1_xml_filename}"))
    vfs.map_file(f"{descriptor_dir}\\{gen2_xml_filename}", absolute_path(f"_data/loaders/hyperv/{gen2_xml_filename}"))
    vfs.map_file(f"{descriptor_dir}\\{gen1_vmcx_filename}", absolute_path(f"_data/loaders/hyperv/{gen1_vmcx_filename}"))
    vfs.map_file(f"{descriptor_dir}\\{gen2_vmcx_filename}", absolute_path(f"_data/loaders/hyperv/{gen2_vmcx_filename}"))

    # Fake disks
    vfs.map_file_fh(f"{disk_dir}\\Default Generation 1.vhdx", BytesIO())
    vfs.map_file_fh(f"{disk_dir}\\Default Generation 2.vhdx", BytesIO())

    gen1_xml_path = vfs.path(f"{descriptor_dir}\\{gen1_xml_filename}")
    gen2_xml_path = vfs.path(f"{descriptor_dir}\\{gen2_xml_filename}")
    gen1_vmcx_path = vfs.path(f"{descriptor_dir}\\{gen1_vmcx_filename}")
    gen2_vmcx_path = vfs.path(f"{descriptor_dir}\\{gen2_vmcx_filename}")

    assert HyperVLoader.detect(gen1_xml_path)
    with mock.patch("dissect.target.container.open") as mocked_open:
        loader = HyperVLoader(gen1_xml_path)
        loader.map(target_bare)

        mocked_open.assert_called_with(vfs.path(f"{disk_dir}\\Default Generation 1.vhdx").resolve())

    assert HyperVLoader.detect(gen2_xml_path)
    with mock.patch("dissect.target.container.open") as mocked_open:
        loader = HyperVLoader(gen2_xml_path)
        loader.map(target_bare)

        mocked_open.assert_called_with(vfs.path(f"{disk_dir}\\Default Generation 2.vhdx").resolve())

    assert HyperVLoader.detect(gen1_vmcx_path)
    with mock.patch("dissect.target.container.open") as mocked_open:
        loader = HyperVLoader(gen1_vmcx_path)
        loader.map(target_bare)

        mocked_open.assert_called_with(vfs.path(f"{disk_dir}\\Default Generation 1.vhdx").resolve())

    assert HyperVLoader.detect(gen2_vmcx_path)
    with mock.patch("dissect.target.container.open") as mocked_open:
        loader = HyperVLoader(gen2_vmcx_path)
        loader.map(target_bare)

        mocked_open.assert_called_with(vfs.path(f"{disk_dir}\\Default Generation 2.vhdx").resolve())



from __future__ import annotations

import argparse
import time
from dataclasses import dataclass
from struct import pack
from typing import Iterator
from unittest.mock import MagicMock, patch

import paho.mqtt.client as mqtt
import pytest

from dissect.target import Target
from dissect.target.loaders.mqtt import Broker, MQTTConnection, case


class MQTTMock(MagicMock):
    disks = []
    hostname = ""

    def fill_disks(self, sizes: list[int]) -> None:
        self.disks = []
        pattern = list(range(8))
        for size in sizes:
            self.disks.append(bytearray(pattern) * 64 * size)  # sizes in sectors of 512

    def publish(self, topic: str, *args) -> None:
        response = MagicMock()
        tokens = topic.split("/")
        command = tokens[2]
        if command == "TOPO":
            tokens[2] = "ID"
            response.topic = "/".join(tokens)
            for host in self.hostnames:
                response.payload = host.encode("utf-8")
                self.on_message(self, None, response)
        elif tokens[2] == "INFO":
            tokens[2] = "DISKS"
            response.topic = "/".join(tokens)
            response.payload = pack("<B", len(self.disks))
            for disk in self.disks:
                response.payload += pack("<IQ", 512, len(disk))
        elif tokens[2] == "SEEK":
            tokens[2] = "READ"
            response.topic = "/".join(tokens)
            begin = int(tokens[4], 16)
            end = int(tokens[5], 16)
            response.payload = self.disks[int(tokens[3])][begin : begin + end]
        else:
            return
        self.on_message(self, None, response)


@dataclass
class MockSeekMessage:
    data: bytes = b""


class MockBroker(MagicMock):
    _seek = False

    def seek(self, *args) -> None:
        self._seek = True

    def read(self, *args) -> MockSeekMessage | None:
        if self._seek:
            self._seek = False
            return MockSeekMessage(data=b"010101")
        return None


@pytest.fixture
def mock_broker() -> Iterator[MockBroker]:
    yield MockBroker()


@pytest.mark.parametrize(
    "alias, hosts, disks, disk, seek, read, expected",
    [
        ("host1", ["host1"], [3], 0, 0, 3, b"\x00\x01\x02"),  # basic
        ("host2", ["host2"], [10], 0, 1, 3, b"\x01\x02\x03"),  # + use offset
        ("group1", ["host3"], [10], 0, 1, 3, b"\x01\x02\x03"),  # + use alias
        ("group2", ["host4"], [10, 10, 1], 1, 1, 3, b"\x01\x02\x03"),  # + use disk 2
        ("group3", ["host4", "host5"], [10, 10, 1], 1, 1, 3, b"\x01\x02\x03"),  # + use disk 2
    ],
)
@patch.object(time, "sleep")  # improve speed during test, no need to wait for peers
def test_remote_loader_stream(
    time: MagicMock,
    monkeypatch: pytest.MonkeyPatch,
    alias: str,
    hosts: list[str],
    disks: list[int],
    disk: int,
    seek: int,
    read: int,
    expected: bytes,
) -> None:
    monkeypatch.setattr(mqtt, "Client", MQTTMock)

    broker = Broker("0.0.0.0", "1884", "key", "crt", "ca", "case1", "user", "pass")
    broker.connect()
    broker.mqtt_client.fill_disks(disks)
    broker.mqtt_client.hostnames = hosts

    with patch("dissect.target.loaders.mqtt.MQTTLoader.broker", broker):
        targets = list(
            Target.open_all(
                [f"mqtt://{alias}?broker=0.0.0.0&port=1884&key=key&crt=crt&ca=ca&peers=1&case=case1"],
            )
        )
        assert len(targets) == len(hosts)
        target = targets[-1]
        target.disks[disk].seek(seek)
        data = target.disks[disk].read(read)
        assert data == expected


def test_mqtt_loader_prefetch(mock_broker: MockBroker) -> None:
    connection = MQTTConnection(mock_broker, "")
    connection.prefetch_factor_inc = 10
    assert connection.factor == 1
    assert connection.prev == -1
    connection.read(1, 0, 100, 0)
    assert connection.factor == 1
    assert connection.prev == 0
    connection.read(1, 100, 100, 0)
    assert connection.factor == connection.prefetch_factor_inc + 1
    assert connection.prev == 100
    connection.read(1, 1200, 100, 0)
    assert connection.factor == (connection.prefetch_factor_inc * 2) + 1
    assert connection.prev == 1200


@pytest.mark.parametrize(
    "case_name, parse_result",
    [
        ("valid_case_name", "valid_case_name"),
        ("ValidCase123", "ValidCase123"),
        ("valid_case_123", "valid_case_123"),
        ("invalid-case", pytest.raises(argparse.ArgumentTypeError)),
        ("invalid case", pytest.raises(argparse.ArgumentTypeError)),
        ("invalid.case", pytest.raises(argparse.ArgumentTypeError)),
        ("invalid@case", pytest.raises(argparse.ArgumentTypeError)),
        ("", pytest.raises(argparse.ArgumentTypeError)),
    ],
    ids=[
        "valid_case_lower",
        "valid_case_mixed",
        "valid_case_with_numbers",
        "invalid_case_hyphen",
        "invalid_case_space",
        "invalid_case_dot",
        "invalid_case_special_char",
        "invalid_case_empty",
    ],
)
def test_case(case_name, parse_result: str | pytest.RaisesContext[argparse.ArgumentTypeError]) -> None:
    if isinstance(parse_result, str):
        assert case(case_name) == parse_result
    else:
        with parse_result:
            case(case_name)



from flow.record.fieldtypes import posix_path

from dissect.target.plugins.os.unix.bsd.osx._os import MacPlugin
from tests._utils import absolute_path


def test_unix_bsd_osx_os(target_osx_users, fs_osx):
    target_osx_users.add_plugin(MacPlugin)

    interface = absolute_path("_data/plugins/os/unix/bsd/osx/_os/en0.plist")
    fs_osx.map_file("/private/var/db/dhcpclient/leases/en0.plist", interface)

    hostname = target_osx_users.hostname
    version = target_osx_users.version
    users = list(target_osx_users.users())
    ips = target_osx_users.ips
    ips.sort()

    dissect_user = users[0]
    test_user = users[1]

    assert hostname == "dummys Mac"
    assert version == "macOS 11.7.5 (20G1225)"

    assert len(users) == 2
    assert len(ips) == 2

    assert dissect_user.name == "_dissect"
    assert dissect_user.passwd == "*"
    assert dissect_user.home == posix_path("/Users/dissect")
    assert dissect_user.shell == "/usr/bin/false"
    assert dissect_user.source == "/var/db/dslocal/nodes/Default/users/_dissect.plist"

    assert test_user.home is None

    assert ips == ["10.42.43.63", "10.42.43.64"]



from pathlib import Path
from unittest.mock import call, patch

from dissect.hypervisor.descriptor.vmx import VMX

from dissect.target.containers.vmdk import VmdkContainer
from dissect.target.loaders.vmwarevm import VmwarevmLoader
from dissect.target.target import Target
from tests._utils import mkdirs


@patch("dissect.target.loaders.vmx.VmdkContainer")
@patch("dissect.target.loaders.vmx.vmx.VMX")
def test_vmwarevm_loader(VMX: VMX, VmdkContainer: VmdkContainer, target_bare: Target, tmp_path: Path):
    root = tmp_path
    mkdirs(root, ["Test.vmwarevm"])
    (root / "Test.vmwarevm" / "Test.vmx").touch()

    VMX.parse.return_value = VMX
    VMX.disks.return_value = ["mock.vmdk"]
    VmdkContainer.return_value = VmdkContainer

    vmwarevm_loader = VmwarevmLoader(root / "Test.vmwarevm")
    vmwarevm_loader.map(target_bare)

    assert len(target_bare.disks) == 1
    assert VmdkContainer.mock_calls == [call(root.resolve() / "Test.vmwarevm" / "mock.vmdk")]



from datetime import datetime, timezone

from dissect.target.plugins.apps.remoteaccess.teamviewer import TeamViewerPlugin
from tests._utils import absolute_path


def test_teamviewer_plugin_global_log(target_win_users, fs_win):
    teamviewer_logfile = absolute_path("_data/plugins/apps/remoteaccess/teamviewer/TestTeamviewer.log")
    target_logfile_name = "/sysvol/Program Files/TeamViewer/TestTeamviewer.log"

    _, _, map_path = target_logfile_name.partition("sysvol/")
    fs_win.map_file(map_path, teamviewer_logfile)

    tvp = TeamViewerPlugin(target_win_users)

    records = list(tvp.logs())
    assert len(records) == 4

    record = records[0]
    assert record.ts == datetime(2021, 11, 11, 12, 34, 56, tzinfo=timezone.utc)
    assert record.message == "Strip the headers, trace the source!"
    assert record.source == target_logfile_name
    assert record.username is None
    assert record.user_id is None
    assert record.user_home is None


def test_teamviewer_plugin_user_log(target_win_users, fs_win):
    teamviewer_logfile = absolute_path("_data/plugins/apps/remoteaccess/teamviewer/TestTeamviewer.log")
    user_details = target_win_users.user_details.find(username="John")
    target_logfile_name = f"{user_details.home_path}/appdata/roaming/teamviewer/teamviewer_TEST_logfile.log"

    _, _, map_path = target_logfile_name.partition("C:/")
    fs_win.map_file(map_path, teamviewer_logfile)

    tvp = TeamViewerPlugin(target_win_users)

    records = list(tvp.logs())
    assert len(records) == 4

    record = records[0]
    assert record.ts == datetime(2021, 11, 11, 12, 34, 56, tzinfo=timezone.utc)
    assert record.message == "Strip the headers, trace the source!"
    assert record.source == target_logfile_name
    assert record.username == user_details.user.name
    assert record.user_id == user_details.user.sid
    assert record.user_home == user_details.user.home


def test_teamviewer_plugin_special_date_parsing(target_win_users, fs_win):
    teamviewer_logfile = absolute_path("_data/plugins/apps/remoteaccess/teamviewer/TestTeamviewer.log")
    user_details = target_win_users.user_details.find(username="John")
    target_logfile_name = f"{user_details.home_path}/appdata/roaming/teamviewer/teamviewer_TEST_logfile.log"

    _, _, map_path = target_logfile_name.partition("C:/")
    fs_win.map_file(map_path, teamviewer_logfile)

    tvp = TeamViewerPlugin(target_win_users)

    records = list(tvp.logs())
    assert len(records) == 4

    record_2 = records[1]
    assert record_2.ts == datetime(2021, 11, 11, 12, 35, 55, tzinfo=timezone.utc)
    assert record_2.message == "Should be year 2021"

    record_3 = records[2]
    assert record_3.ts == datetime(2021, 11, 11, 12, 36, 11, tzinfo=timezone.utc)
    assert record_3.message == "Should discard the milliseconds properly"

    record_4 = records[3]
    assert record_4.ts == datetime(2021, 11, 11, 12, 37, 00, tzinfo=timezone.utc)
    assert record_4.message == "Should be year 2021"



from dissect.target import Target
from dissect.target.plugins.os.unix.linux.android._os import AndroidPlugin


def test_android_os(target_android: Target) -> None:
    target_android.add_plugin(AndroidPlugin)

    assert target_android.os == "android"
    assert target_android.version == "Android 4.4.2 (2013-10-31)"
    assert target_android.hostname == "TMG28071935"



from dissect.evidence import ad1

from dissect.target import filesystem
from dissect.target.loader import Loader


class AD1Loader(Loader):
    """Load AccessData's forensic image format (AD1) files."""

    def __init__(self, path, **kwargs):
        path = path.resolve()

        super().__init__(path)
        self.ad1 = ad1.AD1(path.open("rb"))

    @staticmethod
    def detect(path):
        return path.suffix.lower() == ".ad1"

    def map(self, target):
        pass


class AD1File(filesystem.VirtualFile):
    def open(self):
        return self.entry.open()



import textwrap
from pathlib import Path
from unittest.mock import MagicMock, call, patch

from dissect.target.loaders.utm import UtmLoader
from dissect.target.target import Target
from tests._utils import mkdirs


@patch("dissect.target.loaders.utm.container")
def test_utm_loader(container: MagicMock, target_bare: Target, tmp_path: Path):
    dummy_plist = """<?xml version="1.0" encoding="UTF-8"?>
    <!DOCTYPE plist PUBLIC "-//Apple//DTD PLIST 1.0//EN" "http://www.apple.com/DTDs/PropertyList-1.0.dtd">
    <plist version="1.0">
    <dict>
        <key>Drive</key>
        <array>
            <dict>
                <key>Identifier</key>
                <string>Test</string>
                <key>ImageName</key>
                <string>Test.qcow2</string>
                <key>ImageType</key>
                <string>Disk</string>
                <key>Interface</key>
                <string>IDE</string>
                <key>InterfaceVersion</key>
                <integer>1</integer>
                <key>ReadOnly</key>
                <false/>
            </dict>
        </array>
    </dict>
    </plist>
    """

    root = tmp_path
    mkdirs(root, ["Test.utm"])
    (root / "Test.utm" / "config.plist").write_text(textwrap.dedent(dummy_plist))

    utm_path = root / "Test.utm"
    assert UtmLoader.detect(utm_path)

    utm_loader = UtmLoader(utm_path)
    utm_loader.map(target_bare)

    assert len(target_bare.disks) == 1
    assert container.open.mock_calls == [call(root.resolve() / "Test.utm" / "Data" / "Test.qcow2")]



import re
from datetime import datetime, timezone
from functools import lru_cache
from pathlib import Path
from typing import Iterator

from defusedxml import ElementTree
from flow.record.base import RE_VALID_FIELD_NAME

from dissect.target import plugin
from dissect.target.exceptions import FileNotFoundError as DissectFileNotFoundError
from dissect.target.exceptions import PluginError, UnsupportedPluginError
from dissect.target.helpers.fsutil import has_glob_magic
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugins.apps.webserver.webserver import (
    WebserverAccessLogRecord,
    WebserverPlugin,
)

LOG_RECORD_NAME = "filesystem/windows/iis/logs"

BASIC_RECORD_FIELDS = [
    ("datetime", "ts"),
    ("net.ipaddress", "client_ip"),
    ("net.ipaddress", "server_ip"),
    ("string", "username"),
    ("string", "server_name"),
    ("string", "site_name"),
    ("string", "request_method"),
    ("string", "request_path"),
    ("string", "request_query"),
    ("string", "request_size_bytes"),
    ("string", "response_size_bytes"),
    ("string", "process_time_ms"),
    # https://docs.microsoft.com/en-US/troubleshoot/iis/http-status-code
    ("string", "service_status_code"),
    # https://docs.microsoft.com/en-us/windows/win32/debug/system-error-codes?redirectedfrom=MSDN#system-error-codes-1
    ("string", "win32_status_code"),
    ("string", "log_format"),
    ("string", "log_file"),
]
BasicRecordDescriptor = TargetRecordDescriptor(LOG_RECORD_NAME, BASIC_RECORD_FIELDS)

# Simplified reverse of flow.record.base.RE_VALID_FIELD_NAME
FIELD_NAME_INVALID_CHARS_RE = re.compile(r"[^a-zA-Z0-9]")


class IISLogsPlugin(WebserverPlugin):
    """IIS 7 (and above) logs plugin.

    References:
        - https://docs.microsoft.com/en-us/iis/get-started/planning-your-iis-architecture/introduction-to-applicationhostconfig
        - https://docs.microsoft.com/en-us/previous-versions/iis/6.0-sdk/ms525807%28v=vs.90%29
    """  # noqa: E501

    APPLICATION_HOST_CONFIG = "sysvol/windows/system32/inetsrv/config/applicationHost.config"

    DEFAULT_LOG_PATHS = [
        "sysvol\\Windows\\System32\\LogFiles\\W3SVC*\\*.log",
        "sysvol\\Windows.old\\Windows\\System32\\LogFiles\\W3SVC*\\*.log",
        "sysvol\\inetpub\\logs\\LogFiles\\*.log",
        "sysvol\\inetpub\\logs\\LogFiles\\W3SVC*\\*.log",
        "sysvol\\Resources\\Directory\\*\\LogFiles\\Web\\W3SVC*\\*.log",
    ]

    __namespace__ = "iis"

    def __init__(self, target):
        super().__init__(target)
        self.config = self.target.fs.path(self.APPLICATION_HOST_CONFIG)
        self.log_dirs = self.get_log_dirs()

        self._create_extended_descriptor = lru_cache(4096)(self._create_extended_descriptor)

    def check_compatible(self) -> None:
        if not self.log_dirs:
            raise UnsupportedPluginError("No IIS log files found")

    @plugin.internal
    def get_log_dirs(self) -> list[tuple[str, Path]]:
        log_paths = set()

        if (sysvol_files := self.target.fs.path("sysvol/files")).exists():
            log_paths.add(("auto", sysvol_files))

        try:
            xml_data = ElementTree.fromstring(self.config.open().read(), forbid_dtd=True)
            for log_file_element in xml_data.findall("*/sites/*/logFile"):
                log_format = log_file_element.get("logFormat") or "W3C"
                if log_dir := log_file_element.get("directory"):
                    log_dir = self.target.resolve(log_dir)
                    log_paths.add((log_format, log_dir))

        except (ElementTree.ParseError, DissectFileNotFoundError) as e:
            self.target.log.warning("Error while parsing %s:%s", self.config, e)

        for log_path in self.DEFAULT_LOG_PATHS:
            try:
                # later on we use */*.log to collect the files, so we need to move up 2 levels
                log_dir = self.target.fs.path(log_path).parents[1]
            except IndexError:
                self.target.log.error("Incompatible path found: %s", log_path)
                continue

            if not has_glob_magic(str(log_dir)) and log_dir.exists():
                log_paths.add(("auto", log_dir))
                continue

            for _log_dir_str in self.target.fs.glob(str(log_dir)):
                if not (_log_dir := self.target.fs.path(_log_dir_str)).is_dir():
                    continue
                log_paths.add(("auto", _log_dir))

        return list(log_paths)

    @plugin.internal
    def iter_log_format_path_pairs(self) -> list[tuple[str, str]]:
        for log_format, log_dir_path in self.log_dirs:
            for log_file in log_dir_path.glob("*/*.log"):
                yield (log_format, log_file)

    @plugin.internal
    def parse_autodetect_format_log(self, path: Path) -> Iterator[BasicRecordDescriptor]:
        first_line = path.open().readline().decode("utf-8", errors="backslashreplace").strip()
        if first_line.startswith("#"):
            yield from self.parse_w3c_format_log(path)
        else:
            yield from self.parse_iis_format_log(path)

    @plugin.internal
    def parse_iis_format_log(self, path: Path) -> Iterator[BasicRecordDescriptor]:
        """Parse log file in IIS format and stream log records.

        This format is not the default IIS log format.

        References:
            - https://docs.microsoft.com/en-us/previous-versions/iis/6.0-sdk/ms525807%28v=vs.90%29#iis-log-file-format
            - https://docs.microsoft.com/en-us/previous-versions/windows/it-pro/windows-server-2003/cc728311%28v=ws.10%29
            - https://learn.microsoft.com/en-us/iis/configuration/system.applicationHost/sites/site/logFile
        """  # noqa: E501

        tzinfo = None
        try:
            tzinfo = self.target.datetime.tzinfo
        except PluginError:
            # No target context available
            pass

        def parse_datetime(date_str: str, time_str: str) -> datetime:
            # Example: 10/1/2021 7:19:59
            # "time is recorded as local time." [^3]
            return datetime.strptime(f"{date_str} {time_str}", "%m/%d/%Y %H:%M:%S").replace(tzinfo=tzinfo)

        for line in path.open().readlines():
            # even though the docs say that IIS log format is ASCII format,
            # it is possible to select UTF-8 in configuration
            line = line.decode("utf-8", errors="backslashreplace")

            # Example:
            # 127.0.0.1, -, 9/20/2021, 8:51:21, W3SVC1, DESKTOP-PJOQLJS, 127.0.0.1, 0, 691, 5005, 404, 2, GET, /some, -,
            parts = [part.strip() for part in line.strip().split(",")]

            if len(parts) != 16:
                self.target.log.debug('Unrecognised log line format, skipping: "%s"', line)
                continue

            row = {
                "log_file": str(path),
                "log_format": "IIS",
                "ts": parse_datetime(parts[2], parts[3]),
                "client_ip": parts[0],
                "username": parts[1],
                "site_name": parts[4],
                "server_name": parts[5],
                "server_ip": parts[6],
                "process_time_ms": parts[7],
                "request_size_bytes": parts[8],
                "response_size_bytes": parts[9],
                "service_status_code": parts[10],
                "win32_status_code": parts[11],
                "request_method": parts[12],
                "request_path": parts[13],
                "request_query": parts[14],
                "_target": self.target,
            }
            row = replace_dash_with_none(row)
            yield BasicRecordDescriptor(**row)

    def _create_extended_descriptor(self, extra_fields: tuple[tuple[str, str]]) -> TargetRecordDescriptor:
        return TargetRecordDescriptor(LOG_RECORD_NAME, BASIC_RECORD_FIELDS + list(extra_fields))

    @plugin.internal
    def parse_w3c_format_log(self, path: Path) -> Iterator[TargetRecordDescriptor]:
        """Parse log file in W3C format and stream log records.

        This is the default logging format for IIS [^3].

        References:
            - https://docs.microsoft.com/en-us/previous-versions/iis/6.0-sdk/ms525807%28v=vs.90%29#w3c-extended-log-file-format
            - https://docs.microsoft.com/en-us/previous-versions/windows/it-pro/windows-server-2003/cc786596%28v=ws.10%29
            - https://learn.microsoft.com/en-us/iis/configuration/system.applicationHost/sites/site/logFile
        """  # noqa: E501

        basic_fields = {
            "c-ip",
            "s-ip",
            "cs-username",
            "s-computername",
            "s-sitename",
            "cs-method",
            "cs-uri-stem",
            "cs-uri-query",
            "cs-bytes",
            "sc-bytes",
            "time-taken",
            "sc-status",
            "sc-win32-status",
        }

        record_descriptor = None
        fields = []
        extra_fields = []
        for line in path.open().readlines():
            line = line.decode("utf-8", errors="backslashreplace").strip()

            if line.startswith("#Fields"):
                _, _, fields_str = line.partition("Fields: ")
                fields = fields_str.split()
                extra_fields = sorted(set(fields) - basic_fields)
                extra_fields_with_types = [("string", normalise_field_name(f)) for f in extra_fields]
                record_descriptor = self._create_extended_descriptor(tuple(extra_fields_with_types))
                continue
            elif line.startswith("#"):
                continue

            values = line.split()

            if len(values) != len(fields):
                self.target.log.warning('Log values do not match fields defined, skipping: "%s"', line)
                continue

            if not record_descriptor:
                self.target.log.warning(
                    'Comment line with the fields defined should come before the values, skipping: "%s"', line
                )

            raw = replace_dash_with_none(dict(zip(fields, values)))

            # Example:
            # {
            #     "c-ip": "127.0.0.1",
            #     "cs(Cookie)": null,
            #     "cs(Referer)": null,
            #     "cs(User-Agent)": "Mozilla/5.0+(Windows+NT+10.0;+Win64;+x64)+AppleWebKit/537.36+(KHTML,+like+Gecko)+Chrome/93.0.4577.82+Safari/537.36+Edg/93.0.961.52",  # noqa: E501
            #     "cs-bytes": "714",
            #     "cs-host": "127.0.0.1",
            #     "cs-method": "GET",
            #     "cs-uri-query": null,
            #     "cs-uri-stem": "/nonexistent-path/path+path2",
            #     "cs-username": null,
            #     "cs-version": "HTTP/1.1",
            #     "custom-field-1": null,
            #     "custom-field-2": "Cache-Control:+max-age=0++Connection:+keep-alive++Accept:+text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.9++Accept-Encoding:+gzip,+deflate,+br++Accept-Language:+en-US,en;q=0.9++Host:+127.0.0.1++User-Agent:+Mozilla/5.0+(Windows+NT+10.0;+Win64;+x64)+AppleWebKit/537.36+(KHTML,+like+Gecko)+Chrome/93.0.4577.82+Safari/537.36+Edg/93.0.961.52++sec-ch-ua:+\"Microsoft+Edge\";v=\"93\",+\"+Not;A+Brand\";v=\"99\",+\"Chromium\";v=\"93\"++sec-ch-ua-mobile:+?0++sec-ch-ua-platform:+\"Windows\"++Upgrade-Insecure-Requests:+1++Sec-Fetch-Site:+none++Sec-Fetch-Mode:+navigate++Sec-Fetch-User:+?1++Sec-Fetch-Dest:+document++",  # noqa: E501
            #     "date": "2021-10-01",
            #     "s-computername": "DESKTOP-PJOQLJS",
            #     "s-ip": "127.0.0.1",
            #     "s-port": "80",
            #     "s-sitename": "W3SVC1",
            #     "sc-bytes": "5143",
            #     "sc-status": "404",
            #     "sc-substatus": "11",
            #     "sc-win32-status": "0",
            #     "time": "18:03:57",
            #     "time-taken": "1"
            # }

            # Make the datetime timezone aware.
            # "the time stamp for each W3C Extended Logging log record is UTC-based." [^3]
            ts = None
            if raw.get("date") and raw.get("time"):
                ts = datetime.strptime(f"{raw['date']} {raw['time']}", "%Y-%m-%d %H:%M:%S").replace(tzinfo=timezone.utc)

            yield record_descriptor(
                ts=ts,
                client_ip=raw.get("c-ip"),
                server_ip=raw.get("s-ip"),
                username=raw.get("cs-username"),
                server_name=raw.get("s-computername"),
                site_name=raw.get("s-sitename"),
                request_method=raw.get("cs-method"),
                request_path=raw.get("cs-uri-stem"),
                request_query=raw.get("cs-uri-query"),
                request_size_bytes=raw.get("cs-bytes"),
                response_size_bytes=raw.get("sc-bytes"),
                process_time_ms=raw.get("time-taken"),
                service_status_code=raw.get("sc-status"),
                win32_status_code=raw.get("sc-win32-status"),
                log_format="W3C",
                log_file=str(path),
                _target=self.target,
                **{normalise_field_name(field): raw.get(field) for field in extra_fields},
            )

    @plugin.export(record=BasicRecordDescriptor)
    def logs(self) -> Iterator[TargetRecordDescriptor]:
        """Return contents of IIS (v7 and above) log files.

        Internet Information Services (IIS) for Windows Server is a manageable web server for hosting anything on the
        web. Logs files might, for example, contain traces that indicate that the web server has been exploited.
        Supported log formats: IIS, W3C.
        """
        for log_format, log_file in self.iter_log_format_path_pairs():
            if log_format == "IIS":
                parse_func = self.parse_iis_format_log
            elif log_format == "W3C":
                parse_func = self.parse_w3c_format_log
            elif log_format == "auto":
                parse_func = self.parse_autodetect_format_log
            else:
                raise ValueError(f"Unsupported IIS log format: {log_format}")

            self.target.log.info("Parsing IIS log file %s in %s format", log_file, log_format)
            yield from parse_func(log_file)

    @plugin.export(record=WebserverAccessLogRecord)
    def access(self) -> Iterator[WebserverAccessLogRecord]:
        """Return contents of IIS (v7 and above) log files in unified WebserverAccessLogRecord format.

        See function ``iis.logs`` for more information and more verbose IIS records.
        """

        for iis_record in self.logs():
            yield WebserverAccessLogRecord(
                ts=iis_record.ts,
                remote_user=iis_record.username,
                remote_ip=iis_record.client_ip,
                method=iis_record.request_method,
                uri=iis_record.request_path,
                protocol=getattr(iis_record, "cs_version", None),
                status_code=getattr(iis_record, "service_status_code", None),
                bytes_sent=iis_record.response_size_bytes,
                referer=getattr(iis_record, "cs_referer", None),
                useragent=getattr(iis_record, "cs_user_agent", None),
                source=iis_record.log_file,
                _target=self.target,
            )


def replace_dash_with_none(data: dict) -> dict:
    """Replace "-" placeholder in dict values with None"""
    return {k: (None if v == "-" else v) for k, v in data.items()}


def normalise_field_name(field: str) -> str:
    """Replace all character that are not allowed in the field name by flow.record
    with _, and strip all hanging _ from start / end of the string.
    """
    if RE_VALID_FIELD_NAME.match(field):
        return field

    field = FIELD_NAME_INVALID_CHARS_RE.sub("_", field).strip("_").lower()
    return field



from pathlib import Path

from dissect.target.filesystems.dir import DirectoryFilesystem
from dissect.target.helpers.record import WindowsUserRecord
from dissect.target.loader import Loader
from dissect.target.plugin import OSPlugin, export
from dissect.target.plugins.os.windows.registry import RegistryPlugin


class ProfileLoader(Loader):
    """Load NTUSER.DAT files."""

    def __init__(self, path, **kwargs):
        path = Path(path).resolve()
        super().__init__(path)

    @staticmethod
    def detect(path):
        if path.is_dir() and path.joinpath("NTUSER.DAT").exists():
            return True

        return False

    def map(self, target):
        username = self.path.name

        dfs = DirectoryFilesystem(self.path, case_sensitive=False)
        target.filesystems.add(dfs)
        target.fs.mount(f"sysvol/users/{username}", dfs)

        target._os_plugin = ProfileOSPlugin
        target.add_plugin(RegistryPlugin, check_compatible=False)


class ProfileOSPlugin(OSPlugin):
    @classmethod
    def detect(cls, target):
        return True

    @classmethod
    def create(cls, target, sysvol):
        return cls(target)

    @export(property=True)
    def hostname(self):
        return self.target._generic_name

    @export(property=True)
    def ips(self):
        return []

    @export(property=True)
    def version(self):
        return None

    @export
    def users(self):
        yield WindowsUserRecord(
            sid="0",
            name=self.hostname,
            home=f"sysvol/users/{self.hostname}",
            _target=self.target,
        )

    @export(property=True)
    def os(self):
        return "windows"



import logging
from collections import defaultdict
from typing import BinaryIO, Iterator, Union

from dissect.volume import lvm

from dissect.target.volume import LogicalVolumeSystem, Volume

log = logging.getLogger(__name__)

OPEN_TYPES = (
    "linear",
    "striped",
    "mirror",
    "thin",
)

KNOWN_SKIP_TYPES = (
    "snapshot",
    "thin-pool",
)


class LvmVolumeSystem(LogicalVolumeSystem):
    __type__ = "lvm"

    def __init__(self, fh: Union[BinaryIO, list[BinaryIO]], *args, **kwargs):
        self.lvm = lvm.LVM2(fh)
        super().__init__(fh, *args, **kwargs)

    @classmethod
    def open_all(cls, volumes: list[BinaryIO]) -> Iterator[LogicalVolumeSystem]:
        lvm_pvs = defaultdict(list)

        for vol in volumes:
            if not cls.detect_volume(vol):
                continue

            dev = lvm.LVM2Device(vol)
            if metadata := dev.metadata:
                vg_name = next(key for key, value in metadata.items() if isinstance(value, dict))
                lvm_pvs[vg_name].append(dev)

        for pvs in lvm_pvs.values():
            try:
                yield cls(pvs)
            except Exception:
                continue

    @staticmethod
    def _detect(fh: BinaryIO) -> bool:
        vols = [fh] if not isinstance(fh, list) else fh
        for vol in vols:
            if LvmVolumeSystem.detect_volume(vol):
                return True
        return False

    @staticmethod
    def _detect_volume(fh: BinaryIO) -> bool:
        buf = fh.read(4096)
        return b"LABELONE" in buf

    def _volumes(self) -> Iterator[Volume]:
        num = 1

        for lv_name, lv in self.lvm.volume_group.logical_volumes.items():
            if lv.type not in OPEN_TYPES:
                if lv.type not in KNOWN_SKIP_TYPES:
                    log.debug("Skipping unsupported LVM logical volume type: %s (%s)", lv.type, lv)
                continue

            # When composing a vg-lv name, LVM2 replaces hyphens with double hyphens in the vg and lv names
            # Emulate that here for the volume name
            name = f"{lv.vg.name.replace('-', '--')}-{lv_name.replace('-', '--')}"

            fh = lv.open()
            yield Volume(fh, num, None, fh.size, None, name, raw=lv, vs=self)

            num += 1



import datetime
import sys
from pathlib import Path
from unittest.mock import Mock, patch

import pytest

from dissect.target import Target
from dissect.target.plugins.os.windows.amcache import AmcachePlugin
from dissect.target.plugins.os.windows.log.amcache import AmcacheInstallPlugin
from tests._utils import absolute_path


def test_amcache_new_format(target_win, fs_win):
    amcache_file = absolute_path("_data/plugins/os/windows/amcache/amcache-new.hve")
    fs_win.map_file("windows/appcompat/programs/amcache.hve", amcache_file)

    target_win.add_plugin(AmcachePlugin)

    files = list(target_win.amcache.files())
    programs = list(target_win.amcache.programs())

    applications = list(target_win.amcache.applications())
    application_files = list(target_win.amcache.application_files())
    application_shortcuts = list(target_win.amcache.shortcuts())
    drivers = list(target_win.amcache.drivers())
    containers = list(target_win.amcache.device_containers())

    assert len(files) == 0
    assert len(programs) == 0
    assert len(applications) == 118
    assert len(application_files) == 222
    assert len(application_shortcuts) == 65
    assert len(containers) == 9
    assert len(drivers) == 361


def test_amcache_old_format(target_win, fs_win):
    amcache_file = absolute_path("_data/plugins/os/windows/amcache/amcache-old.hve")
    fs_win.map_file("windows/appcompat/programs/amcache.hve", amcache_file)

    target_win.add_plugin(AmcachePlugin)

    files = list(target_win.amcache.files())
    programs = list(target_win.amcache.programs())

    applications = list(target_win.amcache.applications())
    application_files = list(target_win.amcache.application_files())
    application_shortcuts = list(target_win.amcache.shortcuts())
    drivers = list(target_win.amcache.drivers())
    containers = list(target_win.amcache.device_containers())

    assert len(files) == 69
    assert len(programs) == 0
    assert len(applications) == 326
    assert len(application_files) == 0
    assert len(application_shortcuts) == 0
    assert len(containers) == 16
    assert len(drivers) == 0


def test_amcache_windows_11_applaunches(target_win, fs_win):
    applaunch_file = absolute_path("_data/plugins/os/windows/amcache/PcaAppLaunchDic.txt")
    fs_win.map_file("windows/appcompat/pca/PcaAppLaunchDic.txt", applaunch_file)

    target_win.add_plugin(AmcachePlugin)
    applaunches = list(target_win.amcache.applaunches())

    assert len(applaunches) == 55
    assert applaunches[0].ts == datetime.datetime(2022, 12, 17, 13, 27, 53, 96000, tzinfo=datetime.timezone.utc)
    assert applaunches[0].path == "C:\\ProgramData\\Sophos\\AutoUpdate\\Cache\\sophos_autoupdate1.dir\\su-setup32.exe"


def new_read_key_subkeys(self, key):
    base_values = {
        "AppxPackageFullName": "Microsoft.Microsoft3DViewer_7.2105.4012.0_x64__8wekyb3d8bbwe",
        "AppxPackageRelativeId": "Microsoft.Microsoft3DViewer",
        "BinFileVersion": "7.2105.4012.0",
        "BinProductVersion": "7.2105.4012.0",
        "BinaryType": "pe64_amd64",
        "Language": 0,
        "LinkDate": "05/04/2021 17:43:39",
        "LongPathHash": "3dviewer.exe|40f275349895ac70",
        "LowerCaseLongPath": "c:\\program files\\windowsapps\\microsoft.0_x64__8wekyb3d8bbwe\\3dviewer.exe",
        "Name": "3DViewer.exe",
        "OriginalFileName": "3dviewer.exe",
        "ProductName": "view 3d",
        "ProductVersion": "7.2105.4012.0",
        "ProgramId": "0000df892556c2f7a6b7fa69f7009b5c08cb00000904",
        "Publisher": "microsoft corporation",
        "Size": 19456,
        "Usn": 32259776,
        "Version": "7.2105.4012.0",
    }

    mock_values = []
    for key, value in base_values.items():
        mock_value = Mock()
        mock_value.name = key
        mock_value.value = value
        mock_values.append(mock_value)

    mock_value = Mock()
    mock_value.name = "FileId"
    mock_value.value = self._mock_file_id
    mock_values.append(mock_value)

    mock_entry = Mock()
    mock_entry.timestamp = datetime.datetime(2021, 12, 31)
    mock_entry.values = Mock(return_value=mock_values)

    yield mock_entry


@pytest.mark.parametrize(
    "test_file_id,expected_file_id",
    [
        ("00008e01cdeb9a1c23cee421a647f29c45f67623be97", "8e01cdeb9a1c23cee421a647f29c45f67623be97"),
        ("", None),
        (None, None),
    ],
)
@patch.object(AmcachePlugin, "read_key_subkeys", new_read_key_subkeys)
def test_parse_inventory_application_file(target_win, test_file_id, expected_file_id):
    with patch("dissect.target.plugins.os.windows.amcache.ApplicationFileAppcompatRecord") as mock_record:
        amcache_plugin = AmcachePlugin(target_win)
        amcache_plugin._mock_file_id = test_file_id
        records = list(amcache_plugin.parse_inventory_application_file())

        assert len(records) == 1
        if sys.version_info[:2] < (3, 8):
            call_kwargs = mock_record.call_args[1]
        else:
            call_kwargs = mock_record.call_args.kwargs

        assert call_kwargs.get("digests", None) == [None, expected_file_id, None]


def test_amcache_install_entry(target_win: Target):
    amcache_install_plugin = AmcacheInstallPlugin(target_win)

    amcache_install_plugin.logs = Path(absolute_path("_data/plugins/os/windows/amcache/install"))

    entries = list(amcache_install_plugin.amcache_install())

    created_order = [
        r"SOFTWARE\Microsoft\Windows\CurrentVersion\Uninstall\7-Zip",
        r"C:\Program Files\7-Zip\7-zip.dll",
        r"C:\Program Files\7-Zip\7-zip32.dll",
        r"C:\Program Files\7-Zip\7z.dll",
        r"C:\Program Files\7-Zip\7z.exe",
        r"C:\Program Files\7-Zip\7zFM.exe",
        r"C:\Program Files\7-Zip\7zG.exe",
        r"C:\Program Files\7-Zip\Uninstall.exe",
    ]
    assert len(entries) == 8

    for create, entry in zip(created_order, entries):
        assert str(entry.create) == create
        assert str(entry.path) == r"C:\Users\JohnCena"
        assert str(entry.longname) == r"7z2201-x64.exe"



import logging
import sys
from typing import Any, Dict

import structlog


def custom_obj_renderer(
    logger: structlog.types.WrappedLogger,
    name: str,
    event_dict: structlog.types.EventDict,
) -> Dict[Any, str]:
    """Simple str() serialization for the event dict values for purely aesthetic reasons"""
    return {key: str(value) for key, value in event_dict.items()}


def render_stacktrace_only_in_debug_or_less(
    logger: structlog.types.WrappedLogger,
    name: str,
    event_dict: structlog.types.EventDict,
) -> Dict[Any, str]:
    """
    Render a stack trace of an exception only if `logger` is configured with `DEBUG` or lower level,
    otherwise render `str()` representation of an exception.
    """
    if event_dict.get("exc_info"):
        # If configured logging level is less permissive than DEBUG,
        # do not render full stack trace
        # https://docs.python.org/3/library/logging.html#logging-levels
        if logger.getEffectiveLevel() > logging.DEBUG:
            event_dict.pop("exc_info")
            _, exc, _ = sys.exc_info()
            event_dict["exc"] = str(exc)
    return event_dict


def configure_logging(verbose_value: int, be_quiet: bool, as_plain_text: bool = True):
    """Configure logging level for `dissect` root logger.

    By default, if `verbose_value` is not set (equals 0) and `be_quiet` is False,
    set logging level for `dissect` root logger to `WARNING`.

    If `be_quiet` is set to True, logging level is set to the least noisy `CRITICAL` level.
    """

    renderer = (
        structlog.dev.ConsoleRenderer(colors=True, pad_event=10)
        if as_plain_text
        else structlog.processors.JSONRenderer(sort_keys=True)
    )

    attr_processors = [
        # Add the name of the logger to event dict.
        structlog.stdlib.add_logger_name,
        # Add log level to event dict.
        structlog.stdlib.add_log_level,
        # Add a timestamp in ISO 8601 format.
        structlog.processors.TimeStamper(fmt="iso"),
    ]

    structlog.configure(
        processors=(
            [
                # If log level is too low, abort pipeline and throw away log entry.
                structlog.stdlib.filter_by_level,
            ]
            + attr_processors
            + [
                # Perform %-style formatting.
                structlog.stdlib.PositionalArgumentsFormatter(),
                # If the "stack_info" key in the event dict is true, remove it and
                # render the current stack trace in the "stack" key.
                structlog.processors.StackInfoRenderer(),
                custom_obj_renderer,
                render_stacktrace_only_in_debug_or_less,
                # Wrapping is needed in order to use formatter down the line
                structlog.stdlib.ProcessorFormatter.wrap_for_formatter,
            ]
        ),
        # `wrapper_class` is the bound logger that you get back from
        # get_logger(). This one imitates the API of `logging.Logger`.
        wrapper_class=structlog.stdlib.BoundLogger,
        # `logger_factory` is used to create wrapped loggers that are used for
        # OUTPUT. This one returns a `logging.Logger`.
        logger_factory=structlog.stdlib.LoggerFactory(),
        # Effectively freeze configuration after creating the first bound
        # logger.
        cache_logger_on_first_use=True,
    )

    # warnings issued by the ``warnings`` module will be
    # redirected to the ``py.warnings`` logger
    logging.captureWarnings(True)

    dissect_logger = logging.getLogger("dissect")

    if be_quiet:
        dissect_logger.setLevel(level=logging.CRITICAL)
    elif verbose_value == 0:
        dissect_logger.setLevel(level=logging.WARNING)
    elif verbose_value == 1:
        dissect_logger.setLevel(level=logging.INFO)
    elif verbose_value > 1:
        dissect_logger.setLevel(level=logging.DEBUG)
    else:
        pass

    formatter = structlog.stdlib.ProcessorFormatter(processor=renderer, foreign_pre_chain=attr_processors)

    handler = logging.StreamHandler()
    handler.setFormatter(formatter)
    # set handler on a root logger
    logging.getLogger().handlers = [handler]



from __future__ import annotations

import logging
import os
import traceback
import urllib
from collections import defaultdict
from pathlib import Path
from typing import Any, Callable, Generic, Iterator, Optional, TypeVar, Union

from dissect.target import container, filesystem, loader, plugin, volume
from dissect.target.exceptions import (
    FilesystemError,
    PluginError,
    PluginNotFoundError,
    TargetError,
    TargetPathNotFoundError,
    UnsupportedPluginError,
    VolumeSystemError,
)
from dissect.target.helpers import config
from dissect.target.helpers.fsutil import TargetPath
from dissect.target.helpers.loaderutil import extract_path_info
from dissect.target.helpers.record import ChildTargetRecord
from dissect.target.helpers.utils import StrEnum, parse_path_uri, slugify
from dissect.target.plugins.general import default

log = logging.getLogger(__name__)

FunctionTuple = tuple[plugin.Plugin, Optional[Union[plugin.Plugin, property]]]


class Event(StrEnum):
    INCOMPATIBLE_PLUGIN = "incompatible-plugin"
    REGISTERED_PLUGIN = "registered-plugin"
    FUNC_EXEC = "function-execution"
    FUNC_EXEC_ERROR = "function-execution-error"


def getlogger(target):
    if not log.root.handlers:
        log.setLevel(os.getenv("DISSECT_LOG_TARGET", "CRITICAL"))
    return TargetLogAdapter(log, {"target": target})


class TargetLogAdapter(logging.LoggerAdapter):
    def process(self, msg, kwargs):
        return f"{self.extra['target']}: {msg}", kwargs


class Target:
    """The class that represents the target that you are talking to.

    Targets are the glue that connects the different ``Containers``, ``Loaders``, ``Volumes``
    and ``Filesystems`` together.
    ``Loaders`` are used to map the ``Containers``, ``Volumes`` and ``Filesystems`` of the target
    onto the ``Target`` object.

    The plugins of dissect.target get mapped onto the ``Target`` too.
    They become available as attributes on a ``Target`` object. For example, ``t.hostname``, ``t.evtx()``.
    By executing the plugin function with a target, it will perform the function on itself.

    Args:
        path: The path of a target.
    """

    def __init__(self, path: Union[str, Path] = None):
        # WIP, part of the introduction of URI-style paths.
        # Since pathlib.Path does not support URIs, bigger refactoring
        # is needed in order to fully utilise URI's scheme / path / query
        # for Target configuration.
        self.path_scheme, self.path, self.path_query = parse_path_uri(path)
        if self.path is not None:
            self.path = Path(self.path)

        self.props: dict[Any, Any] = dict()
        self.log = getlogger(self)

        self._name = None
        self._plugins: list[plugin.Plugin] = []
        self._functions: dict[str, FunctionTuple] = {}
        self._loader = None
        self._os = None
        self._os_plugin: type[plugin.OSPlugin] = None
        self._child_plugins: dict[str, plugin.ChildTargetPlugin] = {}
        self._cache = dict()
        self._errors = []
        self._applied = False

        # We do not want to look for config files at the Target path if it is actually a child Target.
        config_paths = [Path.cwd(), Path.home()]
        if not isinstance(path, TargetPath):
            config_paths = [self.path] + config_paths

        try:
            self._config = config.load(config_paths)
        except Exception as e:
            self.log.warning("Error loading config file: %s", self.path)
            self.log.debug("", exc_info=e)
            self._config = config.load(None)  # This loads an empty config.

        # Fill the disks and/or volumes and/or filesystems and apply() will
        # make sure that volumes/filesystems gets filled and the filesystems
        # get auto mounted on top of fs.
        #
        # Except when a volume's guid or volumesystem's serial is unknown (in
        # the case of windows), in which case _init_os() and
        # _os_plugin.create() won't be able to figure out where to mount the
        # filesystem. In that case they should be added and mounted explicitly
        # by the loader.
        #
        # A collection of instances of Container() subclasses
        self.disks = DiskCollection(self)
        # A collection of instances of Volume() subclasses
        self.volumes = VolumeCollection(self)
        # A collection of instances of Filesystem() subclasses
        self.filesystems = FilesystemCollection(self)

        self.fs = filesystem.RootFilesystem(self)

    @classmethod
    def set_event_callback(cls, *, event_type: Optional[Event] = None, event_callback: Callable) -> None:
        """Sets ``event_callbacks`` on a Target class.

        ``event_callbacks`` get used to handle specific events denoted by :class:`Event`.
        This records events related to the target, such as:

        - a plugin gets registered to the target
        - a plugin is incompatible with the target
        - a function succeededs in its execution
        - a function fails in execution

        """
        if not hasattr(cls, "event_callbacks"):
            cls.event_callbacks = defaultdict(set)

        # When `event_type` is not set or is None, the callback is a catch-all callback
        cls.event_callbacks[event_type].add(event_callback)

    def send_event(self, event_type: Event, **kwargs) -> None:
        """Notify event callbacks for the given ``event_type``.

        Each event can have multiple callback methods, it calls all the callbacks that fit the corresponding event type.
        ``None`` is a catch-all method for event callbacks that always get called.

        Args:
            event_type: The type of event
        """
        cls = type(self)
        if not hasattr(cls, "event_callbacks"):
            return

        specific_callbacks = cls.event_callbacks.get(event_type, set())
        catch_all_callbacks = cls.event_callbacks.get(None, set())

        for callback in set.union(specific_callbacks, catch_all_callbacks):
            try:
                callback(self, event_type, **kwargs)
            except Exception:
                self.log.warning(f"Can't send event {event_type} to {callback}", exc_info=True)

    def apply(self) -> None:
        """Resolve all disks, volumes and filesystems and load an operating system on the current ``Target``."""
        self.disks.apply()
        self.volumes.apply()
        self.filesystems.apply()
        self._init_os()
        self._mount_others()
        self._applied = True

    @property
    def _generic_name(self) -> str:
        """Return a generic name for this target."""

        generic_name = (self.path and self.path.name) or None

        if not generic_name:
            # The os() plugin function is only available after a Target.apply(), as
            # only then an _os_plugin is assigned.
            try:
                generic_name = f"Unknown-{self.os}"
            except Exception:
                generic_name = "Unknown"

        return generic_name

    @property
    def name(self) -> str:
        """Return a name for this target.

        The function is guaranteed to give back some name.
        The name will be guaranteed to not have slashes, backslashes and spaces.
        The name won't be guaranteed to be unique.

        Returns:
            The name of a target.
        """
        if self._name:
            return self._name

        target_name = None
        try:
            target_name = self.hostname
        except Exception:
            pass
        finally:
            if not target_name:
                target_name = self._generic_name

        target_name = slugify(target_name)

        # As name() may be called before the Target instance is fully
        # instantiated (before apply() is called), we only save the name once
        # all parts of the Target instance are there, meaning an _os_plugin is
        # present, and the name won't change anymore.
        if self._applied:
            self._name = target_name

        return target_name

    @classmethod
    def open(cls, path: Union[str, Path]) -> Target:
        """Try to find a suitable loader for the given path and load a ``Target`` from it.

        Args:
            path: Path to load the ``Target`` from.

        Returns:
            A Target with a linked :class:`~dissect.target.loader.Loader` object.
        """

        path, parsed_path = extract_path_info(path)

        loader_cls = loader.find_loader(path, parsed_path=parsed_path)
        if loader_cls:
            try:
                loader_instance = loader_cls(path, parsed_path=parsed_path)
            except Exception as e:
                raise TargetError(f"Failed to initiate {loader_cls.__name__} for target {path}: {e}") from e
            return cls._load(path, loader_instance)
        return cls.open_raw(path)

    @classmethod
    def open_raw(cls, path: Union[str, Path]) -> Target:
        """Open a Target with the given path using the :class:`~dissect.target.loaders.raw.RawLoader`.

        Args:
            path: Path to load the Target from.
        """
        if not isinstance(path, Path):
            path = Path(path)

        return cls._load(path, loader.RawLoader(path))

    @classmethod
    def open_all(cls, paths: list[Union[str, Path]], include_children: bool = False) -> Iterator[Target]:
        """Yield targets from a list of paths.

        If the path is a directory, iterate files one directory deep.

        Args:
            paths: A list of paths to load ``Targets`` from.

        Raises:
            TargetError: Raised when not a single ``Target`` can be loaded.
        """

        def _find(find_path: Path, parsed_path: Optional[urllib.parse.ParseResult]):
            yield find_path
            if parsed_path is not None:
                return

            if find_path.is_dir():
                yield from find_path.iterdir()

        at_least_one_loaded = False
        fallback_loaders = [loader.DirLoader, loader.RawLoader]

        # Treat every path as a unique target spec
        for path in paths:
            loaded = False
            path, parsed_path = extract_path_info(path)

            # Search for targets one directory deep
            for entry in _find(path, parsed_path):
                loader_cls = loader.find_loader(entry, parsed_path=parsed_path, fallbacks=fallback_loaders)
                if not loader_cls:
                    continue

                getlogger(entry).debug("Attempting to use loader: %s", loader_cls)
                for sub_entry in loader_cls.find_all(entry, parsed_path=parsed_path):
                    try:
                        ldr = loader_cls(sub_entry, parsed_path=parsed_path)
                    except Exception as e:
                        message = "Failed to initiate loader: %s"
                        if isinstance(e, TargetPathNotFoundError):
                            message = "%s"

                        getlogger(sub_entry).error(message, e)
                        getlogger(sub_entry).debug("", exc_info=e)
                        continue

                    try:
                        # Attempt to load the target using this loader
                        target = cls._load(sub_entry, ldr)
                        loaded = True
                        at_least_one_loaded = True
                        yield target

                    except Exception as e:
                        getlogger(sub_entry).error("Failed to load target with loader %s", ldr, exc_info=e)

                    if include_children:
                        try:
                            yield from target.open_children()
                        except Exception as e:
                            getlogger(sub_entry).error("Failed to load child target from %s", target, exc_info=e)

                # Found a compatible loader for the top level path, no need to search a level deeper
                # Going deeper could cause unwanted behaviour
                if loaded and entry is path:
                    break

        if not at_least_one_loaded:
            raise TargetError(f"Failed to find any loader for targets: {paths}")

    def _load_child_plugins(self) -> None:
        """Load special :class:`~dissect.target.plugin.ChildTargetPlugin` plugins.

        These plugins inform the ``Target`` how to deal with child targets, such as VMs from a hypervisor.
        Examples of these plugins are:

        - :class:`~dissect.target.plugins.child.esxi.ESXiChildTargetPlugin`
        - :class:`~dissect.target.plugins.child.hyper-v.HyperVChildTargetPlugin`
        """
        if self._child_plugins:
            return

        for plugin_desc in plugin.child_plugins():
            try:
                plugin_cls = plugin.load(plugin_desc)
                child_plugin = plugin_cls(self)
            except PluginError:
                self.log.exception("Failed to load child plugin: %s", plugin_desc["class"])
                continue
            except Exception:
                self.log.exception("Broken child plugin: %s", plugin_desc["class"])
                continue

            try:
                child_plugin.check_compatible()
                self._child_plugins[child_plugin.__type__] = child_plugin
            except PluginError as e:
                self.log.debug("Child plugin reported itself as incompatible: %s (%s)", plugin_desc["class"], e)
            except Exception:
                self.log.exception(
                    "An exception occurred while checking for child plugin compatibility: %s", plugin_desc["class"]
                )

    def open_child(self, child: Union[str, Path]) -> Target:
        """Open a child target.

        Args:
            child: The location of a target within the current ``Target``.

        Returns:
            An opened ``Target`` object of the child target.
        """
        if isinstance(child, str) and child.isdecimal():
            child_num = int(child)
            for child_record in self.list_children():
                if child_num == 0:
                    return Target.open(self.fs.path(child_record.path))
                child_num -= 1
        else:
            return Target.open(self.fs.path(child))

    def open_children(self, recursive: bool = False) -> Iterator[Target]:
        """Open all the child targets on a ``Target``.

        Will open all discovered child targets if the current ``Target`` has them, such as VMs on a hypervisor.

        Args:
            recursive: Whether to check the child ``Target`` for more ``Targets``.

        Returns:
            An iterator of ``Targets``.
        """
        for child in self.list_children():
            try:
                target = self.open_child(child.path)
            except TargetError as e:
                self.log.error("Failed to open child target %s", child, exc_info=e)
                continue

            yield target

            if recursive:
                yield from target.open_children(recursive=recursive)

    def list_children(self) -> Iterator[ChildTargetRecord]:
        """Lists all child targets that compatible :class:`~dissect.target.plugin.ChildTargetPlugin` classes
        can discover.
        """
        self._load_child_plugins()
        for child_plugin in self._child_plugins.values():
            yield from child_plugin.list_children()

    @classmethod
    def _load(cls, path: Union[str, Path], ldr: loader.Loader) -> Target:
        """Internal function that attemps to load a path using a given loader.

        Args:
            path: The path to the target.
            ldr: The loader to use for loading this target.

        Raises:
            TargetError: If it failed to load a target.

        Returns:
            A ``Target`` object with disks, volumes and/or filesystems mapped by the ``ldr`` from the given ``path``.
        """
        target = cls(path)

        try:
            ldr.map(target)
            target._loader = ldr
            target.apply()
            return target
        except Exception as e:
            raise TargetError(f"Failed to load target: {path}") from e

    def _init_os(self) -> None:
        """Internal function that attemps to load an OSPlugin for this target."""
        if self._os_plugin:
            # If self._os_plugin is already assigned, we expect it to be fully
            # configured and possibly already instantiated (if not, it will
            # be), hence no os_plugin.create() is run and no detection is
            # attempted.
            os_plugin = self._os_plugin

            if isinstance(os_plugin, plugin.OSPlugin):
                self._os_plugin = os_plugin.__class__
            elif issubclass(os_plugin, plugin.OSPlugin):
                if fs := os_plugin.detect(self):
                    os_plugin = os_plugin.create(self, fs)

            self._os = self.add_plugin(os_plugin)
            return

        if not len(self.disks) and not len(self.volumes) and not len(self.filesystems):
            raise TargetError(f"Failed to load target. No disks, volumes or filesystems: {self.path}")

        candidates = []

        for plugin_desc in plugin.os_plugins():
            # Subclassed OS Plugins used to also subclass the detection of the
            # parent. This meant that in order for a subclassed OS Plugin to be a
            # candidate for the final OS Plugin, the parent OS Plugin's detection
            # had to succeed. This caused some complications in cases where the
            # filesystem layout needed some tweaks by a subclassed OS Plugin before
            # parent OS Plugin paths made sense.
            #
            # E.g. VyOS /boot/<version>/live-rw -> map to /
            #
            # Now subclassed OS Plugins are on the same detection "layer" as
            # regular OS Plugins, but can still inherit functions.
            self.log.debug("Loading OS plugin: %s", plugin_desc["class"])
            try:
                os_plugin = plugin.load(plugin_desc)
                fs = os_plugin.detect(self)
            except PluginError:
                self.log.exception("Failed to load OS plugin: %s", plugin_desc["class"])
                continue
            except Exception:
                self.log.exception("Broken OS plugin: %s", plugin_desc["class"])
                continue

            if not fs:
                continue

            self.log.info("Found compatible OS plugin: %s", plugin_desc["class"])
            candidates.append((plugin_desc, os_plugin, fs))

        fs = None
        os_plugin = default.DefaultPlugin

        if candidates:
            plugin_desc, os_plugin, fs = candidates[0]
            for candidate_plugin_desc, candidate_plugin, candidate_fs in candidates[1:]:
                # More specific OS plugins are considered better candidates
                if len(candidate_plugin.mro()) > len(os_plugin.mro()):
                    plugin_desc, os_plugin, fs = candidate_plugin_desc, candidate_plugin, candidate_fs

            self.log.debug("Selected OS plugin: %s", plugin_desc["class"])
        else:
            # No OS detected
            self.log.warning("Failed to find OS plugin, falling back to default")

        self._os_plugin = os_plugin
        self._os = self.add_plugin(os_plugin.create(self, fs))

    def _mount_others(self) -> None:
        root_fs = self.fs
        counter = 0
        path = "/$fs$/fs0"

        for fs in self.filesystems:
            if fs not in root_fs.mounts.values():
                # determine mount point
                while root_fs.path(path).exists():
                    counter += 1
                    path = f"/$fs$/fs{counter}"

                root_fs.mount(path, fs)

    def add_plugin(
        self,
        plugin_cls: Union[plugin.Plugin, type[plugin.Plugin]],
        check_compatible: bool = True,
    ) -> plugin.Plugin:
        """Add and register a plugin by class.

        Args:
            plugin_cls: The plugin to add and register, this can either be a class or instance. When this is a class,
                        it will be instantiated.
            check_compatible: A flag that determines if we check whether the plugin is compatible with the ``Target``.

        Returns:
            The ``plugin_cls`` instance.

        Raises:
            UnsupportedPluginError: Raised when plugins were found, but they were incompatible
            PluginError: Raised when any other exception occurs while trying to load the plugin.
        """
        self.log.debug("Adding plugin: %s", plugin_cls)

        if not isinstance(plugin_cls, plugin.Plugin):
            try:
                p = plugin_cls(self)
            except PluginError:
                raise
            except Exception as e:
                raise PluginError(f"An exception occurred while trying to initialize a plugin: {plugin_cls}") from e
        else:
            p = plugin_cls

        if not isinstance(p, plugin.Plugin):
            raise PluginError(f"Not a subclass of Plugin: {p}")

        if check_compatible:
            try:
                p.check_compatible()
            except PluginError:
                self.send_event(Event.INCOMPATIBLE_PLUGIN, plugin_cls=plugin_cls)
                raise
            except Exception as e:
                raise UnsupportedPluginError(
                    f"An exception occurred while checking for plugin compatibility: {plugin_cls}"
                ) from e

        self._register_plugin_functions(p)

        return p

    def _register_plugin_functions(self, plugin_inst: plugin.Plugin) -> None:
        """Internal function that registers all the exported functions from a given plugin.

        Args:
            plugin_inst: Instance of a plugin.
        """
        self.send_event(Event.REGISTERED_PLUGIN, plugin_inst=plugin_inst)

        self._plugins.append(plugin_inst)

        if plugin_inst.__namespace__:
            self._functions[plugin_inst.__namespace__] = (plugin_inst, plugin_inst)
        else:
            for func in plugin_inst.__functions__:
                # If we getattr here, property members will be executed, so we do that in __getattr__
                self._functions[func] = (plugin_inst, None)

    def get_function(self, function: str) -> FunctionTuple:
        """Attempt to get a given function.

        If the function is not already registered, look for plugins that export the function and register them.

        Args:
            function: Function name to look for.

        Returns:
            A tuple of the plugin and the corresponding function.

        Raises:
            UnsupportedPluginError: Raised when plugins were found, but they were incompatible
            PluginError: Raised when any other exception occurs while trying to load the plugin.
        """
        if function not in self._functions:
            causes = []

            plugin_desc = None
            for plugin_desc in plugin.lookup(function, self._os_plugin):
                try:
                    plugin_cls = plugin.load(plugin_desc)
                    self.add_plugin(plugin_cls)
                    self.log.debug("Found compatible plugin '%s' for function '%s'", plugin_desc["class"], function)
                    break
                except UnsupportedPluginError as e:
                    self.send_event(Event.INCOMPATIBLE_PLUGIN, plugin_desc=plugin_desc)
                    causes.append(e)
            else:
                if plugin_desc:
                    # In this case we made at least one iteration but it was skipped due incompatibility.
                    # Just take the last known cause for now
                    raise UnsupportedPluginError(
                        f"Unsupported function `{function}` for target with OS plugin {self._os_plugin}",
                        extra=causes[1:] if len(causes) > 1 else None,
                    ) from causes[0] if causes else None

        # We still ended up with no compatible plugins
        if function not in self._functions:
            raise PluginNotFoundError(f"Can't find plugin with function `{function}`")

        p, func = self._functions[function]
        if func is None:
            func = getattr(p.__class__, function)
            if not isinstance(func, property) or (
                isinstance(func, property) and getattr(func.fget, "__persist__", False)
            ):
                # If the persist flag is set on a property, store the property result in the function cache
                # This is so we don't have to evaluate the property again
                func = getattr(p, function)
            self._functions[function] = (p, func)

        return p, func

    def has_function(self, function: str) -> bool:
        """Return whether this Target supports a given function.

        Args:
            function: The function name to look for.

        Returns:
            ``True`` if the function can be found, ``False`` otherwise.
        """
        try:
            self.get_function(function)
            return True
        except PluginError:
            return False

    def __getattr__(self, attr: str) -> Union[plugin.Plugin, Any]:
        """Override of the default __getattr__ so plugins and functions can be called from a ``Target`` object."""
        p, func = self.get_function(attr)

        if isinstance(func, property):
            # If it's a property, execute it and return the result
            try:
                result = func.__get__(p)
                self.send_event(Event.FUNC_EXEC, func=attr)
                return result
            except Exception:
                if not attr.startswith("__"):
                    self.send_event(
                        Event.FUNC_EXEC_ERROR,
                        func=attr,
                        stacktrace=traceback.format_exc(),
                    )
                raise

        return func

    def __dir__(self):
        """Override the default __dir__ to provide autocomplete for things like IPython."""
        funcs = []
        if self._os_plugin:
            funcs = list(self._os_plugin.__functions__)

        for plugin_desc in plugin.plugins(self._os_plugin):
            funcs.extend(plugin_desc["functions"])

        result = set(self.__dict__.keys())
        result.update(self.__class__.__dict__.keys())
        result.update(object.__dict__.keys())
        result.update(funcs)

        return list(result)

    def __repr__(self):
        return f"<Target {self.path}>"


T = TypeVar("T")


class Collection(Generic[T]):
    def __init__(self, target: Target):
        self.target = target
        self.entries: list[T] = []

    def add(self, entry: T) -> None:
        self.entries.append(entry)

    def __getitem__(self, k: str) -> T:
        return self.entries[k]

    def __iter__(self) -> Iterator[T]:
        return iter(self.entries)

    def __len__(self) -> int:
        return len(self.entries)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: {self.entries!r}>"


class DiskCollection(Collection[container.Container]):
    def apply(self) -> None:
        for disk in self.entries:
            # Some LVM configurations (i.e. RAID with the metadata at the end of the disk)
            # may be misidentified as having a valid MBR/GPT on some of the disks
            # To counter this, first check if the disk is part of any LVM configurations that we support
            if not volume.is_lvm_volume(disk):
                try:
                    if not hasattr(disk, "vs") or disk.vs is None:
                        disk.vs = volume.open(disk)
                        self.target.log.debug("Opened volume system: %s on %s", disk.vs, disk)

                    if not len(disk.vs.volumes):
                        raise VolumeSystemError("Volume system has no volumes")

                    for vol in disk.vs.volumes:
                        self.target.volumes.add(vol)
                    continue
                except Exception as e:
                    self.target.log.warning(
                        "Can't identify volume system or no volumes found, adding as raw volume instead: %s", disk
                    )
                    self.target.log.debug("", exc_info=e)

            # Fallthrough case for error and if we're part of a logical volume set
            vol = volume.Volume(disk, 1, 0, disk.size, None, None, disk=disk)
            self.target.volumes.add(vol)


class VolumeCollection(Collection[volume.Volume]):
    def apply(self) -> None:
        # We don't want later additions to modify the todo, so make a copy
        todo = self.entries[:]
        fs_volumes = []
        lvm_volumes = []
        encrypted_volumes = []

        while todo:
            new_volumes = []
            lvm_volumes = []
            encrypted_volumes = []

            for vol in todo:
                if volume.is_lvm_volume(vol):
                    lvm_volumes.append(vol)
                elif volume.is_encrypted(vol):
                    encrypted_volumes.append(vol)
                else:
                    # We could be getting "regular" volume systems out of LVM or encrypted volumes
                    # Try to open each volume as a regular volume system, or add as a filesystem if it fails
                    # There are a few scenarios were we want to discard the opened volume, though
                    #
                    # If the current volume offset is 0 and originates from a "regular" volume system, we're likely
                    # opening a volume system on the same disk again
                    # Sometimes BSD systems are configured this way and an FFS volume "starts" at offset 0
                    #
                    # If we opened an empty volume system, it might also be the case that a filesystem actually
                    # "starts" at offset 0

                    # Regardless of what happens, we want to try to open it as a filesystem later on
                    fs_volumes.append(vol)

                    if vol.offset == 0 and vol.vs and vol.vs.__type__ == "disk":
                        # We are going to re-open a volume system on itself, bail out
                        self.target.log.info("Found volume with offset 0, opening as raw volume instead")
                        continue

                    try:
                        vs = volume.open(vol)
                    except Exception:
                        # If opening a volume system fails, there's likely none, so open as a filesystem instead
                        continue

                    if not len(vs.volumes):
                        # We opened an empty volume system, discard
                        continue

                    self.entries.extend(vs.volumes)
                    new_volumes.extend(vs.volumes)

            self.target.log.debug("LVM volumes found: %s", lvm_volumes)
            self.target.log.debug("Encrypted volumes found: %s", encrypted_volumes)

            for lvm in volume.open_lvm(lvm_volumes):
                self.target.log.debug("Opened LVM: %s", lvm)
                for lv in lvm.volumes:
                    self.add(lv)
                    new_volumes.append(lv)

            for enc_volume in encrypted_volumes:
                for dec_volume in volume.open_encrypted(enc_volume):
                    self.target.log.debug("Encrypted volume opened: %s", enc_volume)
                    self.add(dec_volume)
                    new_volumes.append(dec_volume)

            todo = new_volumes

        mv_fs_volumes = []
        for vol in fs_volumes:
            try:
                if getattr(vol, "fs", None) is None:
                    if filesystem.is_multi_volume_filesystem(vol):
                        mv_fs_volumes.append(vol)
                    else:
                        vol.fs = filesystem.open(vol)
                        self.target.log.debug("Opened filesystem: %s on %s", vol.fs, vol)

                if getattr(vol, "fs", None) is not None:
                    self.target.filesystems.add(vol.fs)
            except FilesystemError as e:
                self.target.log.warning("Can't identify filesystem: %s", vol)
                self.target.log.debug("", exc_info=e)

        for fs in filesystem.open_multi_volume(mv_fs_volumes):
            self.target.filesystems.add(fs)
            for vol in fs.volume:
                vol.fs = fs


class FilesystemCollection(Collection[filesystem.Filesystem]):
    def apply(self) -> None:
        for fs in self.entries:
            for subfs in fs.iter_subfs():
                self.add(subfs)
